from hagelslag.data.ModelOutput import ModelOutput
from hagelslag.util.make_proj_grids import read_arps_map_file, read_ncar_map_file, make_proj_grids
import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter
from scipy.signal import fftconvolve
from scipy.stats import gamma, bernoulli
from scipy.interpolate import interp1d
from scipy.spatial import cKDTree
from skimage.morphology import disk
from netCDF4 import Dataset, date2num, num2date
import os
from glob import glob
from os.path import join, exists
import json
from datetime import timedelta


try:
    from ncepgrib2 import Grib2Encode, dump
    grib_support = True
except ImportError("ncepgrib2 not available"):
    grib_support = False


class EnsembleMemberProduct(object):
    def __init__(self, ensemble_name, model_name, member, run_date, variable, start_date, end_date, path, single_step,
                size_distribution_training_path, watershed_obj, map_file=None, condition_model_name=None, condition_threshold=0.5):
        self.ensemble_name = ensemble_name
        self.model_name = model_name
        self.member = member
        self.run_date = run_date
        self.variable = variable
        self.start_date = start_date
        self.end_date = end_date
        self.times = pd.DatetimeIndex(start=self.start_date, end=self.end_date, freq="1H")
        self.forecast_hours = (self.times - self.run_date).astype('timedelta64[h]').values
        self.path = path
        self.single_step = single_step
        self.size_distribution_training_path = size_distribution_training_path
        self.watershed_obj = watershed_obj
        self.track_forecasts = None
        self.data = None
        self.map_file = map_file
        self.proj_dict = None
        self.grid_dict = None
        self.mapping_data = None
        self.condition_model_name = condition_model_name
        self.condition_threshold = condition_threshold
        self.percentiles = None
        self.num_samples = None
        self.percentile_data = None
        if self.map_file is not None:
            if self.map_file[-3:] == "map":
                self.proj_dict, self.grid_dict = read_arps_map_file(self.map_file)
            else:
                self.proj_dict, self.grid_dict = read_ncar_map_file(self.map_file)
            self.mapping_data = make_proj_grids(self.proj_dict, self.grid_dict)
        self.units = ""
        self.nc_patches = None
        self.hail_forecast_table = None

    def load_data(self, num_samples=1000, percentiles=None):
        """
        Args:
            num_samples: Number of random samples at each grid point
            percentiles: Which percentiles to extract from the random samples

        Returns:
        """
        self.percentiles = percentiles
        self.num_samples = num_samples
        if self.model_name.lower() in ["wrf"]:
            mo = ModelOutput(self.ensemble_name, self.member, self.run_date, self.variable,
                             self.start_date, self.end_date, self.path, self.map_file, self.single_step)
            mo.load_data()
            self.data = mo.data[:]
            if mo.units == "m":
                self.data *= 1000
                self.units = "mm"
            else:
                self.units = mo.units
        else:
            if self.track_forecasts is None:
                self.load_track_data()
            self.units = "mm"
            self.data = np.zeros((self.forecast_hours.size,
                                  self.mapping_data["lon"].shape[0],
                                  self.mapping_data["lon"].shape[1]), dtype=np.float32)
            
            if self.percentiles is not None:
                self.percentile_data = np.zeros([len(self.percentiles)] + list(self.data.shape))
            full_condition_name = "condition_" + self.condition_model_name.replace(" ", "-")
            dist_model_name = "dist" + "_" + self.model_name.replace(" ", "-")
            for track_forecast in self.track_forecasts:
                times = track_forecast["properties"]["times"]
                for s, step in enumerate(track_forecast["features"]):
                    forecast_params = step["properties"][dist_model_name]
                    if self.condition_model_name is not None:
                        condition = step["properties"][full_condition_name]
                    else:
                        condition = None
                    forecast_time = self.run_date + timedelta(hours=times[s])
                    if forecast_time in self.times:
                        t = np.where(self.times == forecast_time)[0][0]
                        mask = np.array(step["properties"]["masks"], dtype=int).ravel()
                        rankings = np.argsort(np.array(step["properties"]["timesteps"]).ravel()[mask==1])
                        i = np.array(step["properties"]["i"], dtype=int).ravel()[mask == 1][rankings]
                        j = np.array(step["properties"]["j"], dtype=int).ravel()[mask == 1][rankings]
                        if rankings.size > 0 and forecast_params[0] > 0.1 and 1 < forecast_params[2] < 100:
                            raw_samples = np.sort(gamma.rvs(forecast_params[0], loc=forecast_params[1],
                                                            scale=forecast_params[2],
                                                            size=(num_samples, rankings.size)),
                                                  axis=1)
                            if self.percentiles is None:
                                samples = raw_samples.mean(axis=0)
                                if condition >= self.condition_threshold:
                                    self.data[t, i, j] = samples
                            else:
                                for p, percentile in enumerate(self.percentiles):
                                    if percentile != "mean":
                                        if condition >= self.condition_threshold:
                                            self.percentile_data[p, t, i, j] = np.percentile(raw_samples, percentile,
                                                                                             axis=0)
                                    else:
                                        if condition >= self.condition_threshold:
                                            self.percentile_data[p, t, i, j] = np.mean(raw_samples, axis=0)
                                samples = raw_samples.mean(axis=0)
                                if condition >= self.condition_threshold:
                                    self.data[t, i, j] = samples

    def load_track_data(self):
        run_date_str = self.run_date.strftime("%Y%m%d")
        print("Load track forecasts {0} {1}".format(self.ensemble_name, run_date_str))
        track_files = sorted(glob(self.path + "/".join([run_date_str, self.member]) + "/*.json"))
        if len(track_files) > 0:
            self.track_forecasts = []
            for track_file in track_files:
                tfo = open(track_file)
                self.track_forecasts.append(json.load(tfo))
                tfo.close()
        else:
            self.track_forecasts = []

    def load_forecast_csv_data(self, csv_path):
        forecast_file = join(csv_path, "hail_forecasts_{0}_{1}_{2}.csv".format(self.ensemble_name,
                                                                    self.member,
                                                                    self.run_date.strftime("%Y%m%d-%H%M")))
        if exists(forecast_file):
            self.hail_forecast_table = pd.read_csv(forecast_file)
        return

    def load_forecast_netcdf_data(self, nc_path):
        nc_file = join(nc_path, "{0}_{1}_{2}_model_patches.nc".format(self.ensemble_name,
                                                                      self.run_date.strftime("%Y%m%d-%H%M"),
                                                                      self.member))
        if exists(nc_file):
            print(nc_file)
            nc_patches = Dataset(nc_file)
            nc_times = pd.DatetimeIndex(num2date(nc_patches.variables["time"][:],
                                             nc_patches.variables["time"].units))
            time_indices = np.in1d(nc_times, self.times)
            self.nc_patches = dict()
            self.nc_patches["time"] = nc_times[time_indices]
            self.nc_patches["forecast_hour"] = nc_patches.variables["time"][time_indices]
            self.nc_patches["obj_values"] = nc_patches.variables[nc_patches.object_variable + "_curr"][time_indices]
            self.nc_patches["masks"] = nc_patches.variables["masks"][time_indices]
            self.nc_patches["i"] = nc_patches.variables["i"][time_indices]
            self.nc_patches["j"] = nc_patches.variables["j"][time_indices]
            nc_patches.close()
        
        return

    def quantile_match(self):
        
        mask_indices = np.where(self.nc_patches["masks"] == 1)
        obj_values = self.nc_patches["obj_values"][mask_indices]
        obj_values = np.array(obj_values)
        percentiles = np.linspace(0.1, 99.9, 100)
        
        try:
            filename = self.size_distribution_training_path + '{0}_{1}_Size_Distribution.csv'.format(self.ensemble_name,
                                                                                                    self.watershed_obj)

            train_period_obj_per_vals = pd.read_csv(filename)
            train_period_obj_per_vals = train_period_obj_per_vals.loc[:,"Values"].values
            per_func = interp1d(train_period_obj_per_vals, percentiles / 100.0, 
                                bounds_error=False, fill_value=(0.1, 99.9))
        except:
            obj_per_vals = np.percentile(obj_values, percentiles)
            per_func = interp1d(obj_per_vals, percentiles / 100.0, bounds_error=False, fill_value=(0.1, 99.9))

        obj_percentiles = np.zeros(self.nc_patches["masks"].shape)
        obj_percentiles[mask_indices] = per_func(obj_values)
        obj_hail_sizes = np.zeros(obj_percentiles.shape)
        model_name = self.model_name.replace(" ", "-")
        self.units = "mm"
        self.data = np.zeros((self.forecast_hours.size,
                              self.mapping_data["lon"].shape[0],
                              self.mapping_data["lon"].shape[1]), dtype=np.float32)
        sh = self.forecast_hours.min()
        for p in range(obj_hail_sizes.shape[0]):
            if self.hail_forecast_table.loc[p, self.condition_model_name.replace(" ", "-") + "_conditionthresh"] > 0.5:
                patch_mask = np.where(self.nc_patches["masks"][p] == 1)
                obj_hail_sizes[p,
                               patch_mask[0],
                               patch_mask[1]] = gamma.ppf(obj_percentiles[p,
                                                                          patch_mask[0],
                                                                          patch_mask[1]],
                                                          self.hail_forecast_table.loc[p,
                                                                                       model_name + "_shape"],
                                                          self.hail_forecast_table.loc[p,
                                                                                       model_name + "_location"],
                                                          self.hail_forecast_table.loc[p,
                                                                                       model_name + "_scale"])
                self.data[self.nc_patches["forecast_hour"][p] - sh,
                          self.nc_patches["i"][p, patch_mask[0], patch_mask[1]],
                          self.nc_patches["j"][p, patch_mask[0], patch_mask[1]]] = obj_hail_sizes[p, patch_mask[0], patch_mask[1]]
        return

    def neighborhood_probability(self, threshold, radius):
        """
        Calculate a probability based on the number of grid points in an area that exceed a threshold.

        Args:
            threshold:
            radius:

        Returns:

        """
        weights = disk(radius, dtype=np.uint8)
        thresh_data = np.zeros(self.data.shape[1:], dtype=np.uint8)
        neighbor_prob = np.zeros(self.data.shape, dtype=np.float32)
        for t in np.arange(self.data.shape[0]):
            thresh_data[self.data[t] >= threshold] = 1
            maximized = fftconvolve(thresh_data, weights, mode="same")
            maximized[maximized > 1] = 1
            maximized[maximized < 1] = 0
            neighbor_prob[t] = fftconvolve(maximized, weights, mode="same")
            thresh_data[:] = 0
        neighbor_prob[neighbor_prob < 1] = 0
        neighbor_prob /= weights.sum()
        return neighbor_prob

    def period_max_neighborhood_probability(self, threshold, radius):
        weights = disk(radius, dtype=np.uint8)
        thresh_data = np.zeros(self.data.shape[1:], dtype=np.uint8)
        thresh_data[self.data.max(axis=0) >= threshold] = 1
        maximized = fftconvolve(thresh_data, weights, mode="same")
        maximized[maximized > 1] = 1
        maximized[maximized < 1] = 0
        neighborhood_prob = fftconvolve(maximized, weights, mode="same")
        neighborhood_prob[neighborhood_prob < 1] = 0
        neighborhood_prob /= weights.sum()
        return neighborhood_prob

    def period_surrogate_severe_prob(self, threshold, radius, sigma, stagger):
        i_grid, j_grid = np.indices(self.data.shape[1:])
        max_data = self.data.max(axis=0)
        max_points = np.array(np.where(max_data >= threshold)).T
        max_tree = cKDTree(max_points)
        stagger_points = np.vstack((i_grid[::stagger, ::stagger].ravel(), j_grid[::stagger, ::stagger].ravel())).T
        valid_stagger_points = np.zeros(stagger_points.shape[0])
        stagger_tree = cKDTree(stagger_points)
        hit_points = np.unique(np.concatenate(max_tree.query_ball_tree(stagger_tree, radius)))
        valid_stagger_points[hit_points] += 1
        surrogate_grid = valid_stagger_points.reshape(i_grid[::stagger, ::stagger].shape)
        surrogate_grid = gaussian_filter(surrogate_grid, sigma)
        return surrogate_grid


    def encode_grib2_percentile(self):
        """
        Encodes member percentile data to GRIB2 format.

        Returns:
            Series of GRIB2 messages
        """
        lscale = 1e6
        grib_id_start = [7, 0, 14, 14, 2]
        gdsinfo = np.array([0, np.product(self.data.shape[-2:]), 0, 0, 30], dtype=np.int32)
        lon_0 = self.proj_dict["lon_0"]
        sw_lon = self.grid_dict["sw_lon"]
        if lon_0 < 0:
            lon_0 += 360
        if sw_lon < 0:
            sw_lon += 360
        gdtmp1 = [1, 0, self.proj_dict['a'], 0, float(self.proj_dict['a']), 0, float(self.proj_dict['b']),
                  self.data.shape[-1], self.data.shape[-2], self.grid_dict["sw_lat"] * lscale,
                  sw_lon * lscale, 0, self.proj_dict["lat_0"] * lscale,
                  lon_0 * lscale,
                  self.grid_dict["dx"] * 1e3, self.grid_dict["dy"] * 1e3, 0b00000000, 0b01000000,
                  self.proj_dict["lat_1"] * lscale,
                  self.proj_dict["lat_2"] * lscale, -90 * lscale, 0]
        pdtmp1 = np.array([1,                # parameter category Moisture
                           31,               # parameter number Hail
                           4,                # Type of generating process Ensemble Forecast
                           0,                # Background generating process identifier
                           31,               # Generating process or model from NCEP
                           0,                # Hours after reference time data cutoff
                           0,                # Minutes after reference time data cutoff
                           1,                # Forecast time units Hours
                           0,                # Forecast time
                           1,                # Type of first fixed surface Ground
                           1,                # Scale value of first fixed surface
                           0,                # Value of first fixed surface
                           1,                # Type of second fixed surface
                           1,                # Scale value of 2nd fixed surface
                           0,                # Value of 2nd fixed surface
                           0,                # Derived forecast type
                           self.num_samples  # Number of ensemble members
                           ], dtype=np.int32)
        grib_objects = pd.Series(index=self.times, data=[None] * self.times.size, dtype=object)
        drtmp1 = np.array([0, 0, 4, 8, 0], dtype=np.int32)
        for t, time in enumerate(self.times):
            time_list = list(self.run_date.utctimetuple()[0:6])
            if grib_objects[time] is None:
                grib_objects[time] = Grib2Encode(0, np.array(grib_id_start + time_list + [2, 1], dtype=np.int32))
                grib_objects[time].addgrid(gdsinfo, gdtmp1)
            pdtmp1[8] = (time.to_pydatetime() - self.run_date).total_seconds() / 3600.0
            data = self.percentile_data[:, t] / 1000.0
            masked_data = np.ma.array(data, mask=data <= 0)
            for p, percentile in enumerate(self.percentiles):
                print("GRIB {3} Percentile {0}. Max: {1} Min: {2}".format(percentile, 
                                                                          masked_data[p].max(), 
                                                                          masked_data[p].min(),
                                                                          time))
                if percentile in range(1, 100):
                    pdtmp1[-2] = percentile
                    grib_objects[time].addfield(6, pdtmp1[:-1], 0, drtmp1, masked_data[p])
                else:
                    pdtmp1[-2] = 0
                    grib_objects[time].addfield(2, pdtmp1, 0, drtmp1, masked_data[p])
        return grib_objects

    def encode_grib2_data(self):
        """
        Encodes member percentile data to GRIB2 format.

        Returns:
            Series of GRIB2 messages
        """
        lscale = 1e6
        grib_id_start = [7, 0, 14, 14, 2]
        gdsinfo = np.array([0, np.product(self.data.shape[-2:]), 0, 0, 30], dtype=np.int32)
        lon_0 = self.proj_dict["lon_0"]
        sw_lon = self.grid_dict["sw_lon"]
        if lon_0 < 0:
            lon_0 += 360
        if sw_lon < 0:
            sw_lon += 360
        gdtmp1 = [1, 0, self.proj_dict['a'], 0, float(self.proj_dict['a']), 0, float(self.proj_dict['b']),
                  self.data.shape[-1], self.data.shape[-2], self.grid_dict["sw_lat"] * lscale,
                  sw_lon * lscale, 0, self.proj_dict["lat_0"] * lscale,
                  lon_0 * lscale,
                  self.grid_dict["dx"] * 1e3, self.grid_dict["dy"] * 1e3, 0b00000000, 0b01000000,
                  self.proj_dict["lat_1"] * lscale,
                  self.proj_dict["lat_2"] * lscale, -90 * lscale, 0]
        pdtmp1 = np.array([1,                # parameter category Moisture
                           31,               # parameter number Hail
                           4,                # Type of generating process Ensemble Forecast
                           0,                # Background generating process identifier
                           31,               # Generating process or model from NCEP
                           0,                # Hours after reference time data cutoff
                           0,                # Minutes after reference time data cutoff
                           1,                # Forecast time units Hours
                           0,                # Forecast time
                           1,                # Type of first fixed surface Ground
                           1,                # Scale value of first fixed surface
                           0,                # Value of first fixed surface
                           1,                # Type of second fixed surface
                           1,                # Scale value of 2nd fixed surface
                           0,                # Value of 2nd fixed surface
                           0,                # Derived forecast type
                           1                 # Number of ensemble members
                           ], dtype=np.int32)
        grib_objects = pd.Series(index=self.times, data=[None] * self.times.size, dtype=object)
        drtmp1 = np.array([0, 0, 4, 8, 0], dtype=np.int32)
        for t, time in enumerate(self.times):
            time_list = list(self.run_date.utctimetuple()[0:6])
            if grib_objects[time] is None:
                grib_objects[time] = Grib2Encode(0, np.array(grib_id_start + time_list + [2, 1], dtype=np.int32))
                grib_objects[time].addgrid(gdsinfo, gdtmp1)
            pdtmp1[8] = (time.to_pydatetime() - self.run_date).total_seconds() / 3600.0
            data = self.data[t] / 1000.0
            data[np.isnan(data)] = 0
            masked_data = np.ma.array(data, mask=data<=0)
            pdtmp1[-2] = 0
            grib_objects[time].addfield(1, pdtmp1, 0, drtmp1, masked_data)
        return grib_objects

    def write_grib2_files(self, grib_objects, path):
        for t, time in enumerate(self.times.to_pydatetime()):
            grib_objects[time].end()
            filename = path + "{0}_{1}_{2}_{3}_{4}.grib2".format(self.ensemble_name,
                                                                 self.member,
                                                                 self.model_name.replace(" ", "-"),
                                                                 self.variable,
                                                                 self.run_date.strftime("%Y%m%d%H") +
                                                                 "f{0:02d}".format(self.forecast_hours[t])
                                                                 )
            fo = open(filename, "wb")
            fo.write(grib_objects[time].msg)
            fo.close()


class EnsembleProducts(object):
    """
    Loads in individual ensemble members and generates both grid point and neighborhood probabilities from the
    model output.

    Attributes:
        ensemble_name (str): Name of the ensemble prediction system.
        members (list): List of ensemble member names.
        run_date (datetime.datetime): Initial date and time of the model run
        variable (str): Name of model variable being processed
        start_date (datetime.datetime or date string): Date of the initial forecast time step.
        end_date (datetime.datetime or date string): Date of the final forecast time step.
        times (pandas.DatetimeIndex): A sequence of hourly forecast times.
        path (str): Path to the ensemble model files.
        single_step (bool): If True, each forecast timestep is in a separate file.
        data (ndarray[members, times, y, x]): Forecast data. Initially None.
        units (str): The units of the forecast variable
    """
    def __init__(self, ensemble_name, members, run_date, variable, start_date, end_date, path, single_step):
        self.ensemble_name = ensemble_name
        self.members = members
        self.run_date = run_date
        self.variable = variable
        self.start_date = start_date
        self.end_date = end_date
        self.times = pd.DatetimeIndex(start=self.start_date, end=self.end_date, freq="1H")
        self.path = path
        self.single_step = single_step
        self.data = None
        self.units = ""
    
    def load_data(self):
        """
        Loads data from each ensemble member.
        """
        for m, member in enumerate(self.members):
            mo = ModelOutput(self.ensemble_name, member, self.run_date, self.variable,
                             self.start_date, self.end_date, self.path, self.map_file, self.single_step)
            mo.load_data()
            if self.data is None:
                self.data = np.zeros((len(self.members), mo.data.shape[0], mo.data.shape[1], mo.data.shape[2]),
                                     dtype=np.float32)
            if mo.units == "m":
                self.data[m] = mo.data * 1000
                self.units = "mm"
            else:
                self.data[m] = mo.data
            if self.units == "":
                self.units = mo.units
            del mo.data
            del mo

    def point_consensus(self, consensus_type):
        """
        Calculate grid-point statistics across ensemble members.

        Args:
            consensus_type: mean, std, median, max, or percentile_nn

        Returns:
            EnsembleConsensus containing point statistic
        """
        if "mean" in consensus_type:
            consensus_data = np.mean(self.data, axis=0)
        elif "std" in consensus_type:
            consensus_data = np.std(self.data, axis=0)
        elif "median" in consensus_type:
            consensus_data = np.median(self.data, axis=0)
        elif "max" in consensus_type:
            consensus_data = np.max(self.data, axis=0)
        elif "percentile" in consensus_type:
            percentile = int(consensus_type.split("_")[1])
            consensus_data = np.percentile(self.data, percentile, axis=0)
        else:
            consensus_data = np.zeros(self.data.shape[1:])
        consensus = EnsembleConsensus(consensus_data, consensus_type, self.ensemble_name,
                                      self.run_date, self.variable, self.start_date, self.end_date, self.units)
        return consensus

    def point_probability(self, threshold):
        """
        Determine the probability of exceeding a threshold at a grid point based on the ensemble forecasts at
        that point.

        Args:
            threshold: If >= threshold assigns a 1 to member, otherwise 0.

        Returns:
            EnsembleConsensus
        """
        point_prob = np.zeros(self.data.shape[1:])
        for t in range(self.data.shape[1]):
            point_prob[t] = np.where(self.data[:, t] >= threshold, 1.0, 0.0).mean(axis=0)
        return EnsembleConsensus(point_prob, "point_probability", self.ensemble_name,
                                 self.run_date, self.variable + "_{0:0.2f}_{1}".format(threshold,
                                                                                       self.units.replace(" ", "_")),
                                 self.start_date, self.end_date, "")

    def neighborhood_probability(self, threshold, radius, sigmas=None):
        """
        Hourly probability of exceeding a threshold based on model values within a specified radius of a point.

        Args:
            threshold (float): probability of exceeding this threshold
            radius (int): distance from point in number of grid points to include in neighborhood calculation.
            sigmas (array of ints): Radii for Gaussian filter used to smooth neighborhood probabilities.

        Returns:
            list of EnsembleConsensus objects containing neighborhood probabilities for each forecast hour.
        """
        if sigmas is None:
            sigmas = [0]
        weights = disk(radius)
        filtered_prob = []
        for sigma in sigmas:
            filtered_prob.append(EnsembleConsensus(np.zeros(self.data.shape[1:], dtype=np.float32),
                                                   "neighbor_prob_r_{0:d}_s_{1:d}".format(radius, sigma),
                                                   self.ensemble_name,
                                                   self.run_date, self.variable + "_{0:0.2f}".format(threshold),
                                                   self.start_date, self.end_date, ""))
        thresh_data = np.zeros(self.data.shape[2:], dtype=np.uint8)
        neighbor_prob = np.zeros(self.data.shape[2:], dtype=np.float32)
        for t in range(self.data.shape[1]):
            for m in range(self.data.shape[0]):
                thresh_data[self.data[m, t] >= threshold] = 1
                maximized = fftconvolve(thresh_data, weights, mode="same")
                maximized[maximized > 1] = 1
                maximized[maximized < 1] = 0
                neighbor_prob += fftconvolve(maximized, weights, mode="same")
                neighbor_prob[neighbor_prob < 1] = 0
                thresh_data[:] = 0
            neighbor_prob /= (self.data.shape[0] * float(weights.sum()))
            for s, sigma in enumerate(sigmas):
                if sigma > 0:
                    filtered_prob[s].data[t] = gaussian_filter(neighbor_prob, sigma=sigma)
                else:
                    filtered_prob[s].data[t] = neighbor_prob
            neighbor_prob[:] = 0
        return filtered_prob

    def period_max_neighborhood_probability(self, threshold, radius, sigmas=None):
        """
        Calculates the neighborhood probability of exceeding a threshold at any time over the period loaded.

        Args:
            threshold (float): splitting threshold for probability calculatations
            radius (int): distance from point in number of grid points to include in neighborhood calculation.
            sigmas (array of ints): Radii for Gaussian filter used to smooth neighborhood probabilities.

        Returns:
            list of EnsembleConsensus objects
        """
        if sigmas is None:
            sigmas = [0]
        weights = disk(radius)
        neighborhood_prob = np.zeros(self.data.shape[2:], dtype=np.float32)
        thresh_data = np.zeros(self.data.shape[2:], dtype=np.uint8)
        for m in range(self.data.shape[0]):
            thresh_data[self.data[m].max(axis=0) >= threshold] = 1
            maximized = fftconvolve(thresh_data, weights, mode="same")
            maximized[maximized > 1] = 1
            neighborhood_prob += fftconvolve(maximized, weights, mode="same")
        neighborhood_prob[neighborhood_prob < 1] = 0
        neighborhood_prob /= (self.data.shape[0] * float(weights.sum()))
        consensus_probs = []
        for sigma in sigmas:
            if sigma > 0:
                filtered_prob = gaussian_filter(neighborhood_prob, sigma=sigma)
            else:
                filtered_prob = neighborhood_prob
            ec = EnsembleConsensus(filtered_prob,
                                   "neighbor_prob_{0:02d}-hour_r_{1:d}_s_{2:d}".format(self.data.shape[1],
                                                                                       radius, sigma),
                                   self.ensemble_name,
                                   self.run_date, self.variable + "_{0:0.2f}".format(float(threshold)),
                                   self.start_date, self.end_date, "")
            consensus_probs.append(ec)
        return consensus_probs


class MachineLearningEnsembleProducts(EnsembleProducts):
    """
    Subclass of EnsembleProducts that processes forecasts from machine learning models. In particular, this
    class converts object distribution forecasts to grid point values for each ensemble member.
    """
    def __init__(self, ml_model_name, members, run_date, variable, start_date, end_date, grid_shape, forecast_bins,
                 forecast_json_path, condition_model_name=None, map_file=None):
        self.track_forecasts = {}
        self.grid_shape = grid_shape
        self.forecast_bins = forecast_bins
        self.condition_model_name = condition_model_name
        self.percentile = None
        if map_file is not None:
            if map_file[-3:] == "map":
                self.proj_dict, self.grid_dict = read_arps_map_file(map_file)
            else:
                self.proj_dict, self.grid_dict = read_ncar_map_file(map_file)
        else:
            self.proj_dict = None
            self.grid_dict = None
        super(MachineLearningEnsembleProducts, self).__init__(ml_model_name, members, run_date, variable,
                                                              start_date, end_date, forecast_json_path,
                                                              single_step=False)

    def load_track_forecasts(self):
        """
        Load the track forecasts from each geoJSON file.

        """
        run_date_str = self.run_date.strftime("%Y%m%d")
        print("Load track forecasts {0} {1}".format(self.ensemble_name, run_date_str))
        for member in self.members:
            self.track_forecasts[member] = []
            track_files = sorted(glob(self.path + "/".join([run_date_str, member]) + "/*.json"))
            if len(track_files) > 0:
                self.track_forecasts[member] = []
                for track_file in track_files:
                    tfo = open(track_file)
                    self.track_forecasts[member].append(json.load(tfo))
                    tfo.close()
                    del tfo
        return
    
    def load_data(self, grid_method="gamma", num_samples=1000, condition_threshold=0.5, zero_inflate=False,
                  percentile=None):
        """
        Reads the track forecasts and converts them to grid point values based on random sampling.

        Args:
            grid_method: "gamma" by default
            num_samples: Number of samples drawn from predicted pdf
            condition_threshold: Objects are not written to the grid if condition model probability is below this
                threshold.
            zero_inflate: Whether to sample zeros from a Bernoulli sampler based on the condition model probability
            percentile: If None, outputs the mean of the samples at each grid point, otherwise outputs the specified
                percentile from 0 to 100.

        Returns:
            0 if tracks are successfully sampled on to grid. If no tracks are found, returns -1.
        """
        self.percentile = percentile
        if self.track_forecasts == {}:
            self.load_track_forecasts()
        if self.track_forecasts == {}:
            return -1
        if self.data is None:
            self.data = np.zeros((len(self.members), self.times.size, self.grid_shape[0], self.grid_shape[1]),
                                 dtype=np.float32)
        else:
            self.data[:] = 0
        if grid_method in ["mean", "median", "samples"]:
            for m, member in enumerate(self.members):
                print("Sampling " + member)
                for track_forecast in self.track_forecasts[member]:
                    times = track_forecast["properties"]["times"]
                    for s, step in enumerate(track_forecast["features"]):
                        forecast_pdf = np.array(step['properties'][self.variable + "_" +
                                                                   self.ensemble_name.replace(" ", "-")])
                        forecast_time = self.run_date + timedelta(hours=times[s])
                        t = np.where(self.times == forecast_time)[0][0]
                        mask = np.array(step['properties']["masks"], dtype=int)
                        i = np.array(step['properties']["i"], dtype=int)
                        i = i[mask == 1]
                        j = np.array(step['properties']["j"], dtype=int)
                        j = j[mask == 1]
                        if grid_method == "samples":
                            intensities = np.array(step["properties"]["timesteps"], dtype=float)[mask == 1]
                            rankings = np.argsort(intensities)
                            samples = np.random.choice(self.forecast_bins, size=intensities.size, replace=True,
                                                       p=forecast_pdf)
                            self.data[m, t, i[rankings], j[rankings]] = samples
                        else:
                            if grid_method == "mean":
                                forecast_value = np.sum(forecast_pdf * self.forecast_bins)
                            elif grid_method == "median":
                                forecast_cdf = np.cumsum(forecast_pdf)
                                forecast_value = self.forecast_bins[np.argmin(np.abs(forecast_cdf - 0.5))]
                            else:
                                forecast_value = 0
                            self.data[m, t, i, j] = forecast_value
        if grid_method in ["gamma"]:
            full_condition_name = "condition_" + self.condition_model_name.replace(" ", "-")
            dist_model_name = self.variable + "_" + self.ensemble_name.replace(" ", "-")
            for m, member in enumerate(self.members):
                for track_forecast in self.track_forecasts[member]:
                    times = track_forecast["properties"]["times"]
                    for s, step in enumerate(track_forecast["features"]):
                        forecast_params = step["properties"][dist_model_name]
                        if self.condition_model_name is not None:
                            condition = step["properties"][full_condition_name]
                        else:
                            condition = None
                        forecast_time = self.run_date + timedelta(hours=times[s])
                        if forecast_time in self.times:
                            t = np.where(self.times == forecast_time)[0][0]
                            mask = np.array(step["properties"]["masks"], dtype=int)
                            rankings = np.argsort(step["properties"]["timesteps"])[mask == 1]
                            i = np.array(step["properties"]["i"], dtype=int)[mask == 1][rankings]
                            j = np.array(step["properties"]["j"], dtype=int)[mask == 1][rankings]
                            if rankings.size > 0:
                                raw_samples = np.sort(gamma.rvs(forecast_params[0], loc=forecast_params[1],
                                                                scale=forecast_params[2],
                                                                size=(num_samples, rankings.size)),
                                                      axis=1)
                                if zero_inflate:
                                    raw_samples *= bernoulli.rvs(condition,
                                                                 size=(num_samples, rankings.size))
                                if percentile is None:
                                    samples = raw_samples.mean(axis=0)
                                else:
                                    samples = np.percentile(raw_samples, percentile, axis=0)
                                if condition is None or condition >= condition_threshold:
                                    self.data[m, t, i, j] = samples
        return 0

    def write_grib2(self, path):
        """
        Writes data to grib2 file. Currently, grib codes are set by hand to hail.

        Args:
            path: Path to directory containing grib2 files.

        Returns:

        """
        if self.percentile is None:
            var_type = "mean"
        else:
            var_type = "p{0:02d}".format(self.percentile)
        lscale = 1e6
        grib_id_start = [7, 0, 14, 14, 2]
        gdsinfo = np.array([0, np.product(self.data.shape[-2:]), 0, 0, 30], dtype=np.int32)
        lon_0 = self.proj_dict["lon_0"]
        sw_lon = self.grid_dict["sw_lon"]
        if lon_0 < 0:
            lon_0 += 360
        if sw_lon < 0:
            sw_lon += 360
        gdtmp1 = np.array([7, 1, self.proj_dict['a'], 1, self.proj_dict['a'], 1, self.proj_dict['b'],
                           self.data.shape[-2], self.data.shape[-1], self.grid_dict["sw_lat"] * lscale,
                           sw_lon * lscale, 0, self.proj_dict["lat_0"] * lscale,
                           lon_0 * lscale,
                           self.grid_dict["dx"] * 1e3, self.grid_dict["dy"] * 1e3, 0,
                           self.proj_dict["lat_1"] * lscale,
                           self.proj_dict["lat_2"] * lscale, 0, 0], dtype=np.int32)
        pdtmp1 = np.array([1, 31, 2, 0, 116, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 192, 0, self.data.shape[0]], dtype=np.int32)
        for m, member in enumerate(self.members):
            pdtmp1[-2] = m
            for t, time in enumerate(self.times):
                time_list = list(time.utctimetuple()[0:6])
                grbe = Grib2Encode(0, np.array(grib_id_start + time_list + [2, 1], dtype=np.int32))
                grbe.addgrid(gdsinfo, gdtmp1)
                pdtmp1[8] = (time.to_pydatetime() - self.run_date).total_seconds() / 3600.0
                drtmp1 = np.array([0, 0, 4, 8, 0], dtype=np.int32)
                data = self.data[m, t].astype(np.float32) / 1000.0
                masked_data = np.ma.array(data, mask=data <= 0)
                grbe.addfield(1, pdtmp1, 0, drtmp1, masked_data)
                grbe.end()
                filename = path + "{0}_{1}_mlhail_{2}_{3}.grib2".format(self.ensemble_name.replace(" ", "-"), member,
                                                                        var_type,
                                                                        time.to_datetime().strftime("%Y%m%d%H%M"))
                print("Writing to " + filename)
                grib_file = open(filename, "wb")
                grib_file.write(grbe.msg)
                grib_file.close()
        return


class EnsembleConsensus(object):
    """
    Stores data and metadata for an ensemble consensus product such as a neighborhood probability or an ensemble
    mean or max. Allows for the product to be output to a netCDF file.
    """
    def __init__(self, data, consensus_type, ensemble_name, run_date, variable, start_date, end_date, units):
        self.data = data
        self.data[self.data < 0.001] = 0
        self.consensus_type = consensus_type
        self.ensemble_name = ensemble_name
        self.run_date = run_date
        self.variable = variable
        self.start_date = start_date
        self.end_date = end_date
        self.times = pd.DatetimeIndex(start=start_date, end=end_date, freq="1H")
        self.units = units

    def init_file(self, filename, time_units="seconds since 1970-01-01T00:00"):
        """
        Initializes netCDF file for writing

        Args:
            filename: Name of the netCDF file
            time_units: Units for the time variable in format "<time> since <date string>"
        Returns:
            Dataset object
        """
        if os.access(filename, os.R_OK):
            out_data = Dataset(filename, "r+")
        else:
            out_data = Dataset(filename, "w")
            if len(self.data.shape) == 2:
                for d, dim in enumerate(["y", "x"]):
                    out_data.createDimension(dim, self.data.shape[d])
            else:
                for d, dim in enumerate(["y", "x"]):
                    out_data.createDimension(dim, self.data.shape[d+1])
            out_data.createDimension("time", len(self.times))
            time_var = out_data.createVariable("time", "i8", ("time",))
            time_var[:] = date2num(self.times.to_pydatetime(), time_units)
            time_var.units = time_units
            out_data.Conventions = "CF-1.6"
        return out_data

    def write_to_file(self, out_data):
        """
        Outputs data to a netCDF file. If the file does not exist, it will be created. Otherwise, additional variables
        are appended to the current file

        Args:
            out_data: Full-path and name of output netCDF file
        """
        full_var_name = self.consensus_type + "_" + self.variable
        if "-hour" in self.consensus_type:
            if full_var_name not in out_data.variables.keys():
                var = out_data.createVariable(full_var_name, "f4", ("y", "x"), zlib=True, 
                                              least_significant_digit=3, shuffle=True)
            else:
                var = out_data.variables[full_var_name]
            var.coordinates = "y x"
        else:
            if full_var_name not in out_data.variables.keys():
                var = out_data.createVariable(full_var_name, "f4", ("time", "y", "x"), zlib=True,
                                              least_significant_digit=3, shuffle=True)
            else:
                var = out_data.variables[full_var_name]
            var.coordinates = "time y x"
        var[:] = self.data
        var.units = self.units
        var.long_name = self.consensus_type + "_" + self.variable
        return




import numpy as np
from .ModelGrid import ModelGrid
from datetime import timedelta
from os.path import join


class NCARStormEventModelGrid(ModelGrid):
    """
    Loads model output from the NCAR MMM 1 and 3 km WRF runs on Cheyenne.

    """
    def __init__(self, run_date, variable, start_date, end_date, path):
        self.pressure_levels = np.array([1000, 925, 850, 700, 600, 500, 400, 300, 250, 200, 150, 100])
        self.path = path
        self.forecast_hours = np.arange((start_date - run_date).total_seconds() / 3600,
                                        (end_date - run_date).total_seconds() / 3600 + 1, dtype=int)
        filenames = []
        for hour in self.forecast_hours:
            valid_time = run_date + timedelta(hours=int(hour))
            filename = join(path, run_date.strftime("%Y%m%d%H"),
                            "diags_d01_{0}.nc".format(valid_time.strftime("%Y-%m-%d_%H_%M_%S")))
            filenames.append(filename)
        super(NCARStormEventModelGrid, self).__init__(filenames, run_date, start_date, end_date, variable)

    def format_var_name(self, variable, var_list):
        z_index = None
        if variable in var_list:
            var_name = variable
        elif "_PL" in variable:
            var_parts = variable.split("_")
            var_name = "_".join(var_parts[:-1])
            p_level = int(var_parts[-1])
            z_index = np.where(self.pressure_levels == p_level)[0][0]
        else:
            raise KeyError("{0} not found in {1}".format(variable, ", ".join(var_list)))
        return var_name, z_index



import numpy as np
import pandas as pd
from .ContingencyTable import ContingencyTable

__author__ = "David John Gagne <djgagne@ou.edu>"


class ROC(object):
    def __init__(self, forecasts, observations, thresholds, obs_threshold):
        self.forecasts = forecasts
        self.observations = observations
        self.thresholds = thresholds
        self.obs_threshold = obs_threshold
        self.pod = np.zeros(thresholds.shape)
        self.pofd = np.zeros(thresholds.shape)
        self.far = np.zeros(thresholds.shape)
        self.calc_roc()

    def calc_roc(self):
        ct = ContingencyTable(0, 0, 0, 0)
        for t, threshold in enumerate(self.thresholds):
            tp = np.count_nonzero((self.forecasts >= threshold) & (self.observations >= self.obs_threshold))
            fp = np.count_nonzero((self.forecasts >= threshold) & (self.observations < self.obs_threshold))
            fn = np.count_nonzero((self.forecasts < threshold) & (self.observations >= self.obs_threshold))
            tn = np.count_nonzero((self.forecasts < threshold) & (self.observations < self.obs_threshold))
            ct.update(tp, fp, fn, tn)
            self.pod[t] = ct.pod()
            self.pofd[t] = ct.pofd()
            self.far[t] = ct.far()

    def auc(self):
        return -np.trapz(self.pod, self.pofd)


class DistributedROC(object):
    """
    ROC sparse representation that can be aggregated and can generate ROC curves and performance diagrams.

    A DistributedROC object is given a specified set of thresholds (could be probability or real-valued) and then
    stores a pandas DataFrame of contingency tables for each threshold. The contingency tables are updated with a
    set of forecasts and observations, but the original forecast and observation values are not kept. DistributedROC
    objects can be combined by adding them together or by storing them in an iterable and summing the contents of the
    iterable together. This is especially useful when verifying large numbers of cases in parallel.

    Attributes:
        thresholds (numpy.ndarray): List of probability thresholds in increasing order.
        obs_threshold (float):  Observation values >= obs_threshold are positive events.
        contingency_tables (pandas.DataFrame): Stores contingency table counts for each probability threshold

    Examples:

        >>> import numpy as np
        >>> forecasts = np.random.random(size=1000)
        >>> obs = np.random.random_integers(0, 1, size=1000)
        >>> roc = DistributedROC(thresholds=np.arange(0, 1.1, 0.1), obs_threshold=1)
        >>> roc.update(forecasts, obs)
        >>> print(roc.auc())
    """
    def __init__(self, thresholds=np.arange(0, 1.1, 0.1), obs_threshold=1.0, input_str=None):
        """
        Initializes the DistributedROC object. If input_str is not None, then the DistributedROC object is
         initialized with the contents of input_str. Otherwise an empty contingency table is created.

        Args:
            thresholds (numpy.array): Array of thresholds in increasing order.
            obs_threshold (float): Split threshold (>= is positive event) (< is negative event)
            input_str (None or str): String containing information for DistributedROC
        """
        self.thresholds = thresholds
        self.obs_threshold = obs_threshold
        self.contingency_tables = pd.DataFrame(np.zeros((thresholds.size, 4), dtype=int),
                                               columns=["TP", "FP", "FN", "TN"])
        if input_str is not None:
            self.from_str(input_str)

    def update(self, forecasts, observations):
        """
        Update the ROC curve with a set of forecasts and observations

        Args:
            forecasts: 1D array of forecast values
            observations: 1D array of observation values.
        """
        for t, threshold in enumerate(self.thresholds):
            tp = np.count_nonzero((forecasts >= threshold) & (observations >= self.obs_threshold))
            fp = np.count_nonzero((forecasts >= threshold) &
                                  (observations < self.obs_threshold))
            fn = np.count_nonzero((forecasts < threshold) &
                                  (observations >= self.obs_threshold))
            tn = np.count_nonzero((forecasts < threshold) &
                                  (observations < self.obs_threshold))
            self.contingency_tables.iloc[t] += [tp, fp, fn, tn]
    
    def clear(self):
        self.contingency_tables.loc[:, :] = 0

    def __add__(self, other):
        """
        Add two DistributedROC objects together and combine their contingency table values.

        Args:
            other: Another DistributedROC object.
        """
        sum_roc = DistributedROC(self.thresholds, self.obs_threshold)
        sum_roc.contingency_tables = self.contingency_tables + other.contingency_tables
        return sum_roc

    def merge(self, other_roc):
        """
        Ingest the values of another DistributedROC object into this one and update the statistics inplace.

        Args:
            other_roc: another DistributedROC object.
        """
        if other_roc.thresholds.size == self.thresholds.size and np.all(other_roc.thresholds == self.thresholds):
            self.contingency_tables += other_roc.contingency_tables
        else:
            print("Input table thresholds do not match.")

    def roc_curve(self):
        """
        Generate a ROC curve from the contingency table by calculating the probability of detection (TP/(TP+FN)) and the
        probability of false detection (FP/(FP+TN)).

        Returns:
            A pandas.DataFrame containing the POD, POFD, and the corresponding probability thresholds.
        """
        pod = self.contingency_tables["TP"].astype(float) / (self.contingency_tables["TP"] +
                                                             self.contingency_tables["FN"])
        pofd = self.contingency_tables["FP"].astype(float) / (self.contingency_tables["FP"] +
                                                              self.contingency_tables["TN"])
        return pd.DataFrame({"POD": pod, "POFD": pofd, "Thresholds": self.thresholds},
                            columns=["POD", "POFD", "Thresholds"])

    def performance_curve(self):
        """
        Calculate the Probability of Detection and False Alarm Ratio in order to output a performance diagram.

        Returns:
            pandas.DataFrame containing POD, FAR, and probability thresholds.
        """
        pod = self.contingency_tables["TP"] / (self.contingency_tables["TP"] + self.contingency_tables["FN"])
        far = self.contingency_tables["FP"] / (self.contingency_tables["FP"] + self.contingency_tables["TP"])
        far[(self.contingency_tables["FP"] + self.contingency_tables["TP"]) == 0] = np.nan
        return pd.DataFrame({"POD": pod, "FAR": far, "Thresholds": self.thresholds},
                            columns=["POD", "FAR", "Thresholds"])

    def auc(self):
        """
        Calculate the Area Under the ROC Curve (AUC).
        """
        roc_curve = self.roc_curve()
        return np.abs(np.trapz(roc_curve['POD'], x=roc_curve['POFD']))

    def max_csi(self):
        """
        Calculate the maximum Critical Success Index across all probability thresholds

        Returns:
            The maximum CSI as a float
        """
        csi = self.contingency_tables["TP"] / (self.contingency_tables["TP"] + self.contingency_tables["FN"] +
                                               self.contingency_tables["FP"])
        return csi.max()

    def max_threshold_score(self, score="ets"):
        cts = self.get_contingency_tables()
        scores = np.array([getattr(ct, score)() for ct in cts])
        return self.thresholds[scores.argmax()], scores.max()

    def get_contingency_tables(self):
        """
        Create an Array of ContingencyTable objects for each probability threshold.

        Returns:
            Array of ContingencyTable objects
        """
        return np.array([ContingencyTable(*ct) for ct in self.contingency_tables.values])

    def __str__(self):
        """
        Output the information within the DistributedROC object to a string.
        """
        out_str = "Obs_Threshold:{0:0.2f}".format(self.obs_threshold) + ";"
        out_str += "Thresholds:" + " ".join(["{0:0.2f}".format(t) for t in self.thresholds]) + ";"
        for col in self.contingency_tables.columns:
            out_str += col + ":" + " ".join(["{0:d}".format(t) for t in self.contingency_tables[col]]) + ";"
        out_str = out_str.rstrip(";")
        return out_str

    def __repr__(self):
        return self.__str__()

    def from_str(self, in_str):
        """
        Read the DistributedROC string and parse the contingency table values from it.

        Args:
            in_str (str): The string output from the __str__ method
        """
        parts = in_str.split(";")
        for part in parts:
            var_name, value = part.split(":")
            if var_name == "Obs_Threshold":
                self.obs_threshold = float(value)
            elif var_name == "Thresholds":
                self.thresholds = np.array(value.split(), dtype=float)
                self.contingency_tables = pd.DataFrame(columns=self.contingency_tables.columns,
                                                       data=np.zeros((self.thresholds.size,
                                                                     self.contingency_tables.columns.size)))
            elif var_name in self.contingency_tables.columns:
                self.contingency_tables[var_name] = np.array(value.split(), dtype=int)


class Reliability(object):
    def __init__(self, forecasts, observations, thresholds, obs_threshold):
        self.forecasts = forecasts
        self.observations = observations
        self.thresholds = thresholds
        self.obs_threshold = obs_threshold
        self.pos_relative_frequency = np.zeros(self.thresholds.shape)
        self.total_relative_frequency = np.zeros(self.thresholds.shape)
        self.calc_reliability_curve()

    def calc_reliability_curve(self):
        pos_frequency = np.zeros(self.thresholds.shape)
        total_frequency = np.zeros(self.thresholds.shape)
        for t, threshold in enumerate(self.thresholds[:-1]):
            pos_frequency[t] = np.count_nonzero((threshold <= self.forecasts) &
                                                (self.forecasts < self.thresholds[t+1]) &
                                                (self.observations > self.obs_threshold))
            total_frequency[t] = np.count_nonzero((threshold <= self.forecasts) &
                                                  (self.forecasts < self.thresholds[t+1]))
            if total_frequency[t] > 0:
                self.pos_relative_frequency[t] = pos_frequency[t] / float(total_frequency[t])
                self.total_relative_frequency[t] = total_frequency / self.forecasts.size
            else:
                self.pos_relative_frequency[t] = np.nan
        self.pos_relative_frequency[-1] = np.nan

    def brier_score(self):
        obs_truth = np.where(self.observations >= self.obs_threshold, 1, 0)
        return np.mean((self.forecasts - obs_truth) ** 2)

    def brier_score_components(self):
        obs_truth = np.where(self.observations >= self.obs_threshold, 1, 0)
        climo_freq = obs_truth.sum() / float(obs_truth.size)
        total_freq = self.total_relative_frequency * self.forecasts.size
        bins = 0.5 * (self.thresholds[0:-1] + self.thresholds[1:])
        pos_rel_freq = np.where(np.isnan(self.pos_relative_frequency), 0, self.pos_relative_frequency)
        reliability = np.mean(total_freq * (bins - pos_rel_freq) ** 2)
        resolution = np.mean(total_freq * (pos_rel_freq - climo_freq) ** 2)
        uncertainty = climo_freq * (1 - climo_freq)
        return reliability, resolution, uncertainty

    def brier_skill_score(self):
        obs_truth = np.where(self.observations >= self.obs_threshold, 1, 0)
        climo_freq = obs_truth.sum() / float(obs_truth.size)
        bs_climo = np.mean((climo_freq - obs_truth) ** 2)
        bs = self.brier_score()
        return 1.0 - bs / bs_climo

    def __str__(self):
        return "Brier Score: {0:0.3f}, Reliability: {1:0.3f}, Resolution: {2:0.3f}, Uncertainty: {3:0.3f}".format(
            tuple([self.brier_score()] + list(self.brier_score_components())))


class DistributedReliability(object):
    """
    A container for the statistics required to generate reliability diagrams and calculate the Brier Score.

    DistributedReliabilty objects accept binary probabilistic forecasts and associated observations. The
    forecasts are then discretized into the different probability bins. The total frequency and the frequency
    of positive events for each probability bin are tracked. The Brier Score, Brier Skill Score, and
    Brier score components can all be derived from this information. Like the DistributedROC object,
    DistributedReliability objects can be summed together, and their contents can be output as a string.

    Attributes:
        thresholds (numpy.ndarray): Array of probability thresholds
        obs_threshold (float): Split value (>=) for determining positive observation events
        frequencies (pandas.DataFrame): Stores the total and positive frequencies for each bin

    Examples:

        >>> forecasts = np.random.random(1000)
        >>> obs = np.random.random_integers(0, 1, 1000)
        >>> rel = DistributedReliability()
        >>> rel.update(forecasts, obs)
        >>> print(rel.brier_score())
    """

    def __init__(self, thresholds=np.arange(0, 1.1, 0.05), obs_threshold=1.0, input_str=None):
        """
        Initialize the DistributedReliability object.

        Args:
            thresholds (numpy.ndarray): Array of probability thresholds
            obs_threshold (float): Split value for observations
            input_str (str): String containing information to initialize the object from a text representation.
        """
        self.thresholds = thresholds
        self.obs_threshold = obs_threshold
        self.frequencies = pd.DataFrame(np.zeros((self.thresholds.size, 2), dtype=int),
                                        columns=["Total_Freq", "Positive_Freq"])
        if input_str is not None:
            self.from_str(input_str)

    def update(self, forecasts, observations):
        """
        Update the statistics with a set of forecasts and observations.

        Args:
            forecasts (numpy.ndarray): Array of forecast probability values
            observations (numpy.ndarray): Array of observation values
        """
        for t, threshold in enumerate(self.thresholds[:-1]):
            self.frequencies.loc[t, "Positive_Freq"] += np.count_nonzero((threshold <= forecasts) &
                                                                         (forecasts < self.thresholds[t+1]) &
                                                                         (observations >= self.obs_threshold))
            self.frequencies.loc[t, "Total_Freq"] += np.count_nonzero((threshold <= forecasts) &
                                                                      (forecasts < self.thresholds[t+1]))
    
    def clear(self):
        self.frequencies.loc[:, :] = 0

    def __add__(self, other):
        """
        Add two DistributedReliability objects together and combine their values.

        Args:
            other: a DistributedReliability object

        Returns:
            A DistributedReliability Object
        """
        sum_rel = DistributedReliability(self.thresholds, self.obs_threshold)
        sum_rel.frequencies = self.frequencies + other.frequencies
        return sum_rel

    def merge(self, other_rel):
        """
        Ingest another DistributedReliability and add its contents to the current object.

        Args:
            other_rel: a Distributed reliability object.
        """
        if other_rel.thresholds.size == self.thresholds.size and np.all(other_rel.thresholds == self.thresholds):
            self.frequencies += other_rel.frequencies
        else:
            print("Input table thresholds do not match.")

    def reliability_curve(self):
        """
        Calculates the reliability diagram statistics. The key columns are Bin_Start and Positive_Relative_Freq

        Returns:
            pandas.DataFrame
        """
        total = self.frequencies["Total_Freq"].sum()
        curve = pd.DataFrame(columns=["Bin_Start", "Bin_End", "Bin_Center",
                                      "Positive_Relative_Freq", "Total_Relative_Freq"])
        curve["Bin_Start"] = self.thresholds[:-1]
        curve["Bin_End"] = self.thresholds[1:]
        curve["Bin_Center"] = 0.5 * (self.thresholds[:-1] + self.thresholds[1:])
        curve["Positive_Relative_Freq"] = self.frequencies["Positive_Freq"] / self.frequencies["Total_Freq"]
        curve["Total_Relative_Freq"] = self.frequencies["Total_Freq"] / total
        return curve

    def brier_score_components(self):
        """
        Calculate the components of the Brier score decomposition: reliability, resolution, and uncertainty.
        """
        rel_curve = self.reliability_curve()
        total = self.frequencies["Total_Freq"].sum()
        climo_freq = float(self.frequencies["Positive_Freq"].sum()) / self.frequencies["Total_Freq"].sum()
        reliability = np.sum(self.frequencies["Total_Freq"] * (rel_curve["Bin_Start"] -
                                                               rel_curve["Positive_Relative_Freq"]) ** 2) / total
        resolution = np.sum(self.frequencies["Total_Freq"] * (rel_curve["Positive_Relative_Freq"] - climo_freq) ** 2) \
                     / total
        uncertainty = climo_freq * (1 - climo_freq)
        return reliability, resolution, uncertainty

    def climatology(self):
        """
        Calculates the sample climatological relative frequency of the event being forecast.

        """
        return float(self.frequencies["Positive_Freq"].sum()) / self.frequencies["Total_Freq"].sum()

    def brier_score(self):
        """
        Calculate the Brier Score
        """
        reliability, resolution, uncertainty = self.brier_score_components()
        return reliability - resolution + uncertainty

    def brier_skill_score(self):
        """
        Calculate the Brier Skill Score
        """
        reliability, resolution, uncertainty = self.brier_score_components()
        return (resolution - reliability) / uncertainty

    def __str__(self):
        out_str = "Obs_Threshold:{0:0.2f}".format(self.obs_threshold) + ";"
        out_str += "Thresholds:" + " ".join(["{0:0.2f}".format(t) for t in self.thresholds]) + ";"
        for col in self.frequencies.columns:
            out_str += col + ":" + " ".join(["{0:d}".format(t) for t in self.frequencies[col]]) + ";"
        out_str = out_str.rstrip(";")
        return out_str

    def __repr__(self):
        return self.__str__()

    def from_str(self, in_str):
        """
        Updates the object attributes with the information contained in the input string

        Args:
            in_str (str): String output by the __str__ method containing all of the attribute values

        """
        parts = in_str.split(";")
        for part in parts:
            var_name, value = part.split(":")
            if var_name == "Obs_Threshold":
                self.obs_threshold = float(value)
            elif var_name == "Thresholds":
                self.thresholds = np.array(value.split(), dtype=float)
                self.frequencies = pd.DataFrame(columns=self.frequencies.columns,
                                                data=np.zeros((self.thresholds.size,
                                                              self.frequencies.columns.size)))
            elif var_name in ["Positive_Freq", "Total_Freq"]:
                self.frequencies[var_name] = np.array(value.split(), dtype=int)


class DistributedCRPS(object):
    """
    A container for the data used to calculate the Continuous Ranked Probability Score.

    Attributes:
        thresholds (numpy.ndarray): Array of the intensity threshold bins
        input_str (str): String containing the information for initializing the object
    """

    def __init__(self, thresholds=np.arange(0, 200.0), input_str=None):
        self.thresholds = thresholds
        crps_columns = ["F_2", "F_O", "O_2", "O"]
        self.errors = pd.DataFrame(columns=crps_columns,
                                   data=np.zeros((len(thresholds), len(crps_columns))), dtype=float)
        self.num_forecasts = 0
        if input_str is not None:
            self.from_str(input_str)

    def update(self, forecasts, observations):
        """
        Update the statistics with forecasts and observations.

        Args:
            forecasts: The discrete Cumulative Distribution Functions of
            observations:
        """
        if len(observations.shape) == 1:
            obs_cdfs = np.zeros((observations.size, self.thresholds.size))
            for o, observation in enumerate(observations):
                obs_cdfs[o, self.thresholds >= observation] = 1
        else:
            obs_cdfs = observations
        self.errors["F_2"] += np.sum(forecasts ** 2, axis=0)
        self.errors["F_O"] += np.sum(forecasts * obs_cdfs, axis=0)
        self.errors["O_2"] += np.sum(obs_cdfs ** 2, axis=0)
        self.errors["O"] += np.sum(obs_cdfs, axis=0)
        self.num_forecasts += forecasts.shape[0]

    def __add__(self, other):
        sum_crps = DistributedCRPS(self.thresholds)
        sum_crps.errors = self.errors + other.errors
        sum_crps.num_forecasts = self.num_forecasts + other.num_forecasts
        return sum_crps

    def merge(self, other_crps):
        if other_crps.thresholds.size == self.thresholds.size and np.all(other_crps.thresholds == self.thresholds):
            self.errors += other_crps.errors
            self.num_forecasts += other_crps.num_forecasts
        else:
            print("ERROR: Input table thresholds do not match.")

    def crps(self):
        """
        Calculates the continuous ranked probability score.
        """
        return np.sum(self.errors["F_2"].values - self.errors["F_O"].values * 2.0 + self.errors["O_2"].values) / \
            (self.thresholds.size * self.num_forecasts)

    def crps_climo(self):
        """
        Calculate the climatological CRPS.
        """
        o_bar = self.errors["O"].values / float(self.num_forecasts)
        crps_c = np.sum(self.num_forecasts * (o_bar ** 2) - o_bar * self.errors["O"].values * 2.0 +
                        self.errors["O_2"].values) / float(self.thresholds.size * self.num_forecasts)
        return crps_c

    def crpss(self):
        """
        Calculate the continous ranked probability skill score from existing data.
        """
        crps_f = self.crps()
        crps_c = self.crps_climo()
        return 1.0 - float(crps_f) / float(crps_c)

    def from_str(self, in_str):
        str_parts = in_str.split(";")
        for part in str_parts:
            var_name, value = part.split(":")
            if var_name == "Thresholds":
                self.thresholds = np.array(value.split(), dtype=float)
                self.errors = pd.DataFrame(data=np.zeros((self.thresholds.size, self.errors.columns.size)),
                                           columns=self.errors.columns)
            elif var_name in self.errors.columns:
                self.errors[var_name] = np.array(value.split(), dtype=float)
            elif var_name == "Num_Forecasts":
                self.num_forecasts = int(value)

    def __str__(self):
        out_str = ""
        out_str += "Thresholds:" + " ".join(["{0:0.2f}".format(t) for t in self.thresholds]) + ";"
        for col in self.errors.columns:
            out_str += col + ":" + " ".join(["{0:0.3f}".format(e) for e in self.errors[col]]) + ";"
        out_str += "Num_Forecasts:{0:d}".format(self.num_forecasts)
        return out_str

    def __repr__(self):
        return self.__str__()


def bootstrap(score_objs, n_boot=1000):
    """
    Given a set of DistributedROC or DistributedReliability objects, this function performs a
    bootstrap resampling of the objects and returns n_boot aggregations of them.

    Args:
        score_objs: A list of DistributedROC or DistributedReliability objects. Objects must have an __add__ method
        n_boot (int): Number of bootstrap samples

    Returns:
        An array of DistributedROC or DistributedReliability
    """
    all_samples = np.random.choice(score_objs, size=(n_boot, len(score_objs)), replace=True)
    return all_samples.sum(axis=1)



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes, InsetPosition


def roc_curve(roc_objs, obj_labels, colors, markers, filename, figsize=(8, 8),
              xlabel="Probability of False Detection",
              ylabel="Probability of Detection",
              title="ROC Curve", ticks=np.arange(0, 1.1, 0.1), dpi=300,
              legend_params=None, bootstrap_sets=None, ci=(2.5, 97.5),
              label_fontsize=14, title_fontsize=16, tick_fontsize=12):
    """
    Plots a set receiver/relative operating characteristic (ROC) curves from DistributedROC objects.

    The ROC curve shows how well a forecast discriminates between two outcomes over a series of thresholds. It
    features Probability of Detection (True Positive Rate) on the y-axis and Probability of False Detection
    (False Alarm Rate) on the x-axis. This plotting function allows one to customize the colors and markers of the
    ROC curves as well as the parameters of the legend and the title.

    Args:
        roc_objs (list): DistributedROC objects being plotted.
        obj_labels (list): Label describing the forecast associated with a DistributedROC object.
        colors (list): List of matplotlib-readable colors (names or hex-values) for each curve.
        markers (list): Matplotlib marker (e.g. *, o, v, etc.) for each curve.
        filename (str): Name of figure file being saved.
        figsize (tuple): (Width, height) of the figure in inches.
        xlabel (str): Label for the x-axis.
        ylabel (str): Label for the y-axis.
        title (str): The title of the figure.
        ticks (numpy.ndarray): Values shown on the x and y axes.
        dpi (int): Figure resolution in dots per inch.
        legend_params (None, dict): Keyword arguments for the formatting of the figure legend.
        bootstrap_sets (list): List of lists of DistributedROC objects that were bootstrap resampled for each model.
        ci (tuple of 2 floats): Quantiles of the edges of the bootstrap confidence intervals ranging from 0 to 100.
        label_fontsize (int): Font size of the x and y axis labels.
        title_fontsize (int): Font size of the title.
        tick_fontsize (int): Font size of the x and y tick labels.

    Examples:

        >>> from hagelslag.evaluation import DistributedROC
        >>> import numpy as np
        >>> forecasts = np.random.random(1000)
        >>> obs = np.random.random_integers(0, 1, 1000)
        >>> roc = DistributedROC()
        >>> roc.update(forecasts, obs)
        >>> roc_curve([roc], ["Random"], ["orange"], ["o"], "random_roc.png")
    """
    if legend_params is None:
        legend_params = dict(loc=4, fontsize=12, framealpha=1, frameon=True)
    plt.figure(figsize=figsize, dpi=dpi)
    plt.plot(ticks, ticks, "k--")
    if bootstrap_sets is not None:
        for b, b_set in enumerate(bootstrap_sets):
            broc_curves = np.dstack([b_roc.roc_curve().values for b_roc in b_set])
            pod_range = np.percentile(broc_curves[:,0], ci, axis=1)
            pofd_range = np.percentile(broc_curves[:, 1], ci, axis=1)
            pod_poly = np.concatenate((pod_range[1], pod_range[0, ::-1]))
            pofd_poly = np.concatenate((pofd_range[0], pofd_range[1, ::-1]))
            pod_poly[np.isnan(pod_poly)] = 0
            pofd_poly[np.isnan(pofd_poly)] = 0
            plt.fill(pofd_poly, pod_poly, alpha=0.5, color=colors[b])
    for r, roc_obj in enumerate(roc_objs):
        roc_data = roc_obj.roc_curve()
        plt.plot(roc_data["POFD"], roc_data["POD"], marker=markers[r], color=colors[r], label=obj_labels[r])
    plt.xlabel(xlabel, fontsize=label_fontsize)
    plt.ylabel(ylabel, fontsize=label_fontsize)
    plt.xticks(ticks, fontsize=tick_fontsize)
    plt.yticks(ticks, fontsize=tick_fontsize)
    plt.title(title, fontsize=title_fontsize)
    plt.legend(**legend_params)
    plt.savefig(filename, dpi=dpi, bbox_inches="tight")
    plt.close()


def performance_diagram(roc_objs, obj_labels, colors, markers, filename, figsize=(8, 8),
                        xlabel="Success Ratio (1-FAR)",
                        ylabel="Probability of Detection", ticks=np.arange(0, 1.1, 0.1),
                        dpi=300, csi_cmap="Blues",
                        csi_label="Critical Success Index", title="Performance Diagram",
                        legend_params=None, bootstrap_sets=None, ci=(2.5, 97.5), label_fontsize=14,
                        title_fontsize=16, tick_fontsize=12):
    """
    Draws a performance diagram from a set of DistributedROC objects.

    A performance diagram is a variation on the ROC curve in which the Probability of False Detection on the
    x-axis has been replaced with the Success Ratio (1-False Alarm Ratio or Precision). The diagram also shows
    the Critical Success Index (CSI or Threat Score) as a series of curved contours, and the frequency bias as
    angled diagonal lines. Points along the 1:1 diagonal are unbiased, and better performing models should appear
    in the upper right corner. The performance diagram is particularly useful for displaying verification for
    severe weather warnings as it displays all three commonly used statistics (POD, FAR, and CSI) simultaneously
    on the same chart.

    Args:
        roc_objs (list): DistributedROC objects being plotted.
        obj_labels: list or array of labels describing each DistributedROC object.
        obj_labels (list): Label describing the forecast associated with a DistributedROC object.
        colors (list): List of matplotlib-readable colors (names or hex-values) for each curve.
        markers (list): Matplotlib marker (e.g. *, o, v, etc.) for each curve.
        filename (str): Name of figure file being saved.
        figsize (tuple): (Width, height) of the figure in inches.
        xlabel (str): Label for the x-axis.
        ylabel (str): Label for the y-axis.
        title (str): The title of the figure.
        ticks (numpy.ndarray): Values shown on the x and y axes.
        dpi (int): Figure resolution in dots per inch.
        csi_cmap (str): Matplotlib colormap used to fill CSI contours.
        csi_label (str): Label for CSI colormap.
        legend_params (None or dict): Keyword arguments for the formatting of the figure legend.
        bootstrap_sets (list): A list of arrays of bootstrapped DistributedROC objects. If not None,
            confidence regions will be plotted.
        ci (tuple): tuple of bootstrap confidence interval percentiles.
        label_fontsize (int): Font size of the x and y axis labels.
        title_fontsize (int): Font size of the title.
        tick_fontsize (int): Font size of the x and y tick labels.

    Examples:

        >>> from hagelslag.evaluation import DistributedROC
        >>> import numpy as np
        >>> forecasts = np.random.random(1000)
        >>> obs = np.random.random_integers(0, 1, 1000)
        >>> roc = DistributedROC()
        >>> roc.update(forecasts, obs)
        >>> performance_diagram([roc], ["Random"], ["orange"], ["o"], "random_performance.png")
    """
    if legend_params is None:
        legend_params = dict(loc=4, fontsize=10, framealpha=1, frameon=True)
    plt.figure(figsize=figsize)
    grid_ticks = np.arange(0, 1.01, 0.01)
    sr_g, pod_g = np.meshgrid(grid_ticks, grid_ticks)
    bias = pod_g / sr_g
    csi = 1.0 / (1.0 / sr_g + 1.0 / pod_g - 1.0)
    csi_contour = plt.contourf(sr_g, pod_g, csi, np.arange(0.1, 1.1, 0.1), extend="max", cmap=csi_cmap)
    b_contour = plt.contour(sr_g, pod_g, bias, [0.5, 1, 1.5, 2, 4], colors="k", linestyles="dashed")
    plt.clabel(b_contour, fmt="%1.1f", manual=[(0.2, 0.9), (0.4, 0.9), (0.6, 0.9), (0.7, 0.7)])
    if bootstrap_sets is not None:
        for b, b_set in enumerate(bootstrap_sets):
            perf_curves = np.dstack([b_roc.performance_curve().values for b_roc in b_set])
            pod_range = np.nanpercentile(perf_curves[:, 0], ci, axis=1)
            sr_range = np.nanpercentile(1 - perf_curves[:, 1], ci, axis=1)
            pod_poly = np.concatenate((pod_range[1], pod_range[0, ::-1]))
            sr_poly = np.concatenate((sr_range[1], sr_range[0, ::-1]))
            pod_poly[np.isnan(pod_poly)] = 0
            sr_poly[np.isnan(sr_poly)] = 1
            plt.fill(sr_poly, pod_poly, alpha=0.5, color=colors[b])
    for r, roc_obj in enumerate(roc_objs):
        perf_data = roc_obj.performance_curve()
        plt.plot(1 - perf_data["FAR"], perf_data["POD"], marker=markers[r], color=colors[r], label=obj_labels[r])
    cbar = plt.colorbar(csi_contour)
    cbar.set_label(csi_label)
    plt.xlabel(xlabel, fontsize=label_fontsize)
    plt.ylabel(ylabel, fontsize=label_fontsize)
    plt.xticks(ticks, fontsize=tick_fontsize)
    plt.yticks(ticks, fontsize=tick_fontsize)
    plt.title(title, fontsize=title_fontsize)
    plt.legend(**legend_params)
    plt.savefig(filename, dpi=dpi, bbox_inches="tight")
    plt.close()


def reliability_diagram(rel_objs, obj_labels, colors, markers, filename, figsize=(8, 8), xlabel="Forecast Probability",
                        ylabel="Observed Relative Frequency", ticks=np.arange(0, 1.05, 0.05), dpi=300, inset_size=1.5,
                        title="Reliability Diagram", legend_params=None, bootstrap_sets=None, ci=(2.5, 97.5)):
    """
    Plot reliability curves against a 1:1 diagonal to determine if probability forecasts are consistent with their
    observed relative frequency.

    Args:
        rel_objs (list): List of DistributedReliability objects.
        obj_labels (list): List of labels describing the forecast model associated with each curve.
        colors (list): List of colors for each line
        markers (list): List of line markers
        filename (str): Where to save the figure.
        figsize (tuple): (Width, height) of the figure in inches.
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        ticks (array): Tick value labels for the x and y axes.
        dpi (int): resolution of the saved figure in dots per inch.
        inset_size (float): Size of inset
        title (str): Title of figure
        legend_params (dict): Keyword arguments for the plot legend.
        bootstrap_sets (list): A list of arrays of bootstrapped DistributedROC objects. If not None,
            confidence regions will be plotted.
        ci (tuple): tuple of bootstrap confidence interval percentiles
    """
    if legend_params is None:
        legend_params = dict(loc=4, fontsize=10, framealpha=1, frameon=True)
    fig, ax = plt.subplots(figsize=figsize)
    plt.plot(ticks, ticks, "k--")
    inset_hist = inset_axes(ax, width=inset_size, height=inset_size, loc=2)
    if bootstrap_sets is not None:
        for b, b_set in enumerate(bootstrap_sets):
            brel_curves = np.dstack([b_rel.reliability_curve().values for b_rel in b_set])
            bin_range = np.percentile(brel_curves[:,0], ci, axis=1)
            rel_range = np.percentile(brel_curves[:, 3], ci, axis=1)
            bin_poly = np.concatenate((bin_range[1], bin_range[0, ::-1]))
            rel_poly = np.concatenate((rel_range[1], rel_range[0, ::-1]))
            bin_poly[np.isnan(bin_poly)] = 0
            rel_poly[np.isnan(rel_poly)] = 0
            plt.fill(bin_poly, rel_poly, alpha=0.5, color=colors[b])
    for r, rel_obj in enumerate(rel_objs):
        rel_curve = rel_obj.reliability_curve()
        ax.plot(rel_curve["Bin_Start"], rel_curve["Positive_Relative_Freq"], color=colors[r], marker=markers[r],
                 label=obj_labels[r])
        inset_hist.semilogy(rel_curve["Bin_Start"], rel_curve["Total_Relative_Freq"], color=colors[r],
                            marker=markers[r])
        inset_hist.set_xlabel("Forecast Probability")
        inset_hist.set_ylabel("Forecast Relative Frequency")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.legend(**legend_params)
    ax.set_title(title)
    plt.savefig(filename, dpi=dpi, bbox_inches="tight")
    plt.close()


def attributes_diagram(rel_objs, obj_labels, colors, markers, filename, figsize=(8, 8), xlabel="Forecast Probability",
                       ylabel="Observed Relative Frequency", ticks=np.arange(0, 1.05, 0.05), dpi=300, 
                       title="Attributes Diagram", legend_params=None, inset_params=None,
                       inset_position=(0.12, 0.72, 0.25, 0.25), bootstrap_sets=None, ci=(2.5, 97.5)):
    """
    Plot reliability curves against a 1:1 diagonal to determine if probability forecasts are consistent with their
    observed relative frequency. Also adds gray areas to show where the climatological probabilities lie and what
    areas result in a positive Brier Skill Score.

    Args:
        rel_objs (list): List of DistributedReliability objects.
        obj_labels (list): List of labels describing the forecast model associated with each curve.
        colors (list): List of colors for each line
        markers (list): List of line markers
        filename (str): Where to save the figure.
        figsize (tuple): (Width, height) of the figure in inches.
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        ticks (array): Tick value labels for the x and y axes.
        dpi (int): resolution of the saved figure in dots per inch.
        title (str): Title of figure
        legend_params (dict): Keyword arguments for the plot legend.
        inset_params (dict): Keyword arguments for the inset axis.
        inset_position (tuple): Position of the inset axis in normalized axes coordinates (left, bottom, width, height)
        bootstrap_sets (list): A list of arrays of bootstrapped DistributedROC objects. If not None,
            confidence regions will be plotted.
        ci (tuple): tuple of bootstrap confidence interval percentiles
    """
    if legend_params is None:
        legend_params = dict(loc=4, fontsize=10, framealpha=1, frameon=True)
    if inset_params is None:
        inset_params = dict(width="25%", height="25%", loc=2, axes_kwargs=dict(axisbg='white'))
    fig, ax = plt.subplots(figsize=figsize)
    plt.plot(ticks, ticks, "k--")
    inset_hist = inset_axes(ax, **inset_params)
    ip = InsetPosition(ax, inset_position)
    inset_hist.set_axes_locator(ip)
    climo = rel_objs[0].climatology()
    no_skill = 0.5 * ticks + 0.5 * climo
    skill_x = [climo, climo, 1, 1, climo, climo, 0, 0, climo]
    skill_y = [climo, 1, 1, no_skill[-1], climo, 0, 0, no_skill[0], climo]
    f = ax.fill(skill_x, skill_y, "0.8")
    f[0].set_zorder(1)
    ax.plot(ticks, np.ones(ticks.shape) * climo, "k--")
    if bootstrap_sets is not None:
        for b, b_set in enumerate(bootstrap_sets):
            brel_curves = np.vstack([b_rel.reliability_curve()["Positive_Relative_Freq"].values for b_rel in b_set])
            rel_range = np.nanpercentile(brel_curves, ci, axis=0)
            fb = ax.fill_between(b_rel.thresholds[:-1], rel_range[1], rel_range[0], alpha=0.5, color=colors[b])
            fb.set_zorder(2)
    for r, rel_obj in enumerate(rel_objs):
        rel_curve = rel_obj.reliability_curve()
        ax.plot(rel_curve["Bin_Start"], rel_curve["Positive_Relative_Freq"], color=colors[r],
                marker=markers[r], label=obj_labels[r])
        inset_hist.semilogy(rel_curve["Bin_Start"] * 100, rel_obj.frequencies["Total_Freq"][:-1], color=colors[r],
                            marker=markers[r])
    inset_hist.set_xlabel("Forecast Probability")
    inset_hist.set_ylabel("Frequency")
    ax.annotate("No Skill", (0.6, no_skill[12]), rotation=22.5)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xticks(ticks)
    ax.set_xticklabels((ticks * 100).astype(int))
    ax.set_yticks(ticks)
    ax.set_yticklabels((ticks * 100).astype(int))
    ax.legend(**legend_params)
    ax.set_title(title)
    plt.savefig(filename, dpi=dpi, bbox_inches="tight")
    plt.close()



import unittest
import numpy as np
from datetime import datetime, timedelta
from hagelslag.data.ModelOutput import ModelOutput
from hagelslag.processing.tracker import label_storm_objects, extract_storm_objects


class TestTracking(unittest.TestCase):
    def setUp(self):
        self.model_path = "testdata/spring2015_unidata/"
        self.ensemble_name = "SSEF"
        self.member ="wrf-s3cn_arw"
        self.run_date = datetime(2015, 6, 4)
        self.variable = "uh_max"
        self.start_hour = 18
        self.end_hour = 24
        self.start_date = self.run_date + timedelta(hours=self.start_hour)
        self.end_date = self.run_date + timedelta(hours=self.end_hour)
        self.model_grid = ModelOutput(self.ensemble_name,
                                 self.member,
                                 self.run_date,
                                 self.variable,
                                 self.start_date,
                                 self.end_date,
                                 self.model_path,
                                 "mapfiles/ssef2015.map",
                                 single_step=False)
        self.model_grid.load_data()
        self.model_grid.load_map_info("mapfiles/ssef2015.map")

    def test_model_grid_loading(self):

        self.assertEquals(self.model_grid.lat.shape, self.model_grid.lon.shape,
                          "Lat and Long grids are not equal shaped.")
        self.assertEquals(self.model_grid.data.shape[1:], self.model_grid.lat.shape,
                          "Data shape does not match map shape.")
        self.assertEquals(self.end_hour - self.start_hour + 1, self.model_grid.data.shape[0],
                          "Number of hours loaded does not match specified hours.")

    def test_object_identification(self):
        min_thresh = 10
        max_thresh = 50
        label_grid = label_storm_objects(self.model_grid.data[0], "hyst",
                                         min_thresh, max_thresh, min_area=2, max_area=100)
        label_points = self.model_grid.data[0][label_grid > 0]
        self.assertGreaterEqual(label_points.min(), min_thresh, "Labeled points include those below minimum threshold")
        storm_objs = extract_storm_objects(label_grid, self.model_grid.data[0], self.model_grid.x,
                                           self.model_grid.y, np.array([0]))
        self.assertEquals(len(storm_objs[0]), label_grid.max(), "Storm objects do not match number of labeled objects")



import numpy as np
import pandas as pd
from netCDF4 import Dataset, num2date
from glob import glob
import argparse
from os.path import join


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--path", help="Path to storm patch files")
    parser.add_argument("-n", "--nprocs", type=int, help="Number of processors")
    parser.add_argument("-o", "--out", help="Output path")
    args = parser.parse_args()
    patch_files = sorted(glob(join(args.path, "*.nc")))
    patch_center_list = [get_storm_centers(patch_file) for patch_file in patch_files]
    patch_center_frame = pd.concat(patch_center_list, ignore_index=True)
    patch_center_frame.to_csv(args.out, index=False)

def get_storm_centers(patch_file):
    patch_info = patch_file.split("/")[-1][:-3].split("_")
    run_date = pd.Timestamp(patch_info[-3][:-2])
    member = patch_info[-1]
    patch_data = None
    print(run_date, member)
    with Dataset(patch_file) as patch_set:
       num_patches = patch_set.variables["p"].size
       lons = patch_set.variables["longitude"][:, 32, 32]
       lats = patch_set.variables["latitude"][:, 32, 32]
       valid_date = num2date(patch_set.variables["valid_date"][:], patch_set.variables["valid_date"].units)
       patch_data = pd.DataFrame(dict(longitude=lons, 
                                      latitude=lats, 
                                      member=[member] * num_patches, 
                                      run_date=[run_date] * num_patches,
                                      valid_date=valid_date), columns=["run_date", "member", "valid_date", "longitude", "latitude"])
    return patch_data

if __name__ == "__main__":
    main()



import numpy as np


def melting_layer_height(height_surface, height_700mb, height_500mb, temperature_700mb, temperature_500mb):
    melting_layer_height_asl = -temperature_500mb * (height_700mb - height_500mb) / \
                               (temperature_700mb - temperature_500mb) + height_500mb
    return melting_layer_height_asl - height_surface


def relative_humidity_pressure_level(temperature, specific_humidity, pressure):
    saturation_vapor_pressure = 611.2 * np.exp((17.67 * temperature) / (temperature + 243.5))
    vapor_pressure = specific_humidity / 1000.0 * pressure / 0.622
    return vapor_pressure / saturation_vapor_pressure * 100.0


from .STObject import STObject
from .EnhancedWatershedSegmenter import EnhancedWatershed
from .Hysteresis import Hysteresis
from hagelslag.processing.ObjectMatcher import ObjectMatcher
from scipy.ndimage import find_objects, center_of_mass, gaussian_filter
import numpy as np


def label_storm_objects(data, method, min_intensity, max_intensity, min_area=1, max_area=100, max_range=1,
                        increment=1, gaussian_sd=0):
    """
    From a 2D grid or time series of 2D grids, this method labels storm objects with either the Enhanced Watershed
    or Hysteresis methods.

    Args:
        data: the gridded data to be labeled. Should be a 2D numpy array in (y, x) coordinate order or a 3D numpy array
            in (time, y, x) coordinate order
        method: "ew" or "watershed" for Enhanced Watershed or "hyst" for hysteresis
        min_intensity: Minimum intensity threshold for gridpoints contained within any objects
        max_intensity: For watershed, any points above max_intensity are considered as the same value as max intensity.
            For hysteresis, all objects have to contain at least 1 pixel that equals or exceeds this value
        min_area: (default 1) The minimum area of any object in pixels.
        max_area: (default 100) The area threshold in pixels at which the enhanced watershed ends growth. Object area
            may exceed this threshold if the pixels at the last watershed level exceed the object area.
        max_range: Maximum difference between the maximum and minimum value in an enhanced watershed object before
            growth is stopped.
        increment: Discretization increment for the enhanced watershed
        gaussian_sd: Standard deviation of Gaussian filter applied to data
    Returns:
        label_grid: an ndarray with the same shape as data in which each pixel is labeled with a positive integer value.
    """
    if method.lower() in ["ew", "watershed"]:
        labeler = EnhancedWatershed(min_intensity, increment, max_intensity, max_area, max_range)
    else:
        labeler = Hysteresis(min_intensity, max_intensity)
    if len(data.shape) == 2:
        label_grid = labeler.label(gaussian_filter(data, gaussian_sd))
        label_grid[data < min_intensity] = 0
        if min_area > 1:
            label_grid = labeler.size_filter(label_grid, min_area)
    else:
        label_grid = np.zeros(data.shape, dtype=int)
        for t in range(data.shape[0]):
            label_grid[t] = labeler.label(gaussian_filter(data[t], gaussian_sd))
            label_grid[t][data[t] < min_intensity] = 0
            if min_area > 1:
                label_grid[t] = labeler.size_filter(label_grid[t], min_area)
    return label_grid


def extract_storm_objects(label_grid, data, x_grid, y_grid, times, dx=1, dt=1, obj_buffer=0):
    """
    After storms are labeled, this method extracts the storm objects from the grid and places them into STObjects.
    The STObjects contain intensity, location, and shape information about each storm at each timestep.

    Args:
        label_grid: 2D or 3D array output by label_storm_objects.
        data: 2D or 3D array used as input to label_storm_objects.
        x_grid: 2D array of x-coordinate data, preferably on a uniform spatial grid with units of length.
        y_grid: 2D array of y-coordinate data.
        times: List or array of time values, preferably as integers
        dx: grid spacing in same units as x_grid and y_grid.
        dt: period elapsed between times
        obj_buffer: number of extra pixels beyond bounding box of object to store in each STObject

    Returns:
        storm_objects: list of lists containing STObjects identified at each time.
    """
    storm_objects = []
    if len(label_grid.shape) == 3:
        ij_grid = np.indices(label_grid.shape[1:])
        for t, time in enumerate(times):
            storm_objects.append([])
            object_slices = list(find_objects(label_grid[t], label_grid[t].max()))
            if len(object_slices) > 0:
                for o, obj_slice in enumerate(object_slices):
                    if obj_buffer > 0:
                        obj_slice_buff = [slice(np.maximum(0, osl.start - obj_buffer),
                                                np.minimum(osl.stop + obj_buffer, label_grid.shape[l + 1]))
                                          for l, osl in enumerate(obj_slice)]
                    else:
                        obj_slice_buff = obj_slice
                    storm_objects[-1].append(STObject(data[t][obj_slice_buff],
                                                      np.where(label_grid[t][obj_slice_buff] == o + 1, 1, 0),
                                                      x_grid[obj_slice_buff],
                                                      y_grid[obj_slice_buff],
                                                      ij_grid[0][obj_slice_buff],
                                                      ij_grid[1][obj_slice_buff],
                                                      time,
                                                      time,
                                                      dx=dx,
                                                      step=dt))
                    if t > 0:
                        dims = storm_objects[-1][-1].timesteps[0].shape
                        storm_objects[-1][-1].estimate_motion(time, data[t - 1], dims[1], dims[0])
    else:
        ij_grid = np.indices(label_grid.shape)
        storm_objects.append([])
        object_slices = list(find_objects(label_grid, label_grid.max()))
        if len(object_slices) > 0:
            for o, obj_slice in enumerate(object_slices):
                if obj_buffer > 0:
                    obj_slice_buff = [slice(np.maximum(0, osl.start - obj_buffer),
                                            np.minimum(osl.stop + obj_buffer, label_grid.shape[l + 1]))
                                      for l, osl in enumerate(obj_slice)]
                else:
                    obj_slice_buff = obj_slice
                storm_objects[-1].append(STObject(data[obj_slice_buff],
                                                  np.where(label_grid[obj_slice_buff] == o + 1, 1, 0),
                                                  x_grid[obj_slice_buff],
                                                  y_grid[obj_slice_buff],
                                                  ij_grid[0][obj_slice_buff],
                                                  ij_grid[1][obj_slice_buff],
                                                  times,
                                                  times,
                                                  dx=dx,
                                                  step=dt))
    return storm_objects


def extract_storm_patches(label_grid, data, x_grid, y_grid, times, dx=1, dt=1, patch_radius=16):
    """
    After storms are labeled, this method extracts boxes of equal size centered on each storm from the grid and places
    them into STObjects. The STObjects contain intensity, location, and shape information about each storm
    at each timestep.

    Args:
        label_grid: 2D or 3D array output by label_storm_objects.
        data: 2D or 3D array used as input to label_storm_objects.
        x_grid: 2D array of x-coordinate data, preferably on a uniform spatial grid with units of length.
        y_grid: 2D array of y-coordinate data.
        times: List or array of time values, preferably as integers
        dx: grid spacing in same units as x_grid and y_grid.
        dt: period elapsed between times
        patch_radius: Number of grid points from center of mass to extract

    Returns:
        storm_objects: list of lists containing STObjects identified at each time.
    """
    storm_objects = []
    if len(label_grid.shape) == 3:
        ij_grid = np.indices(label_grid.shape[1:])
        for t, time in enumerate(times):
            storm_objects.append([])
            # object_slices = find_objects(label_grid[t], label_grid[t].max())
            centers = list(center_of_mass(data[t], labels=label_grid[t], index=np.arange(1, label_grid[t].max() + 1)))
            if len(centers) > 0:
                for o, center in enumerate(centers):
                    int_center = np.round(center).astype(int)
                    obj_slice_buff = [slice(int_center[0] - patch_radius, int_center[0] + patch_radius),
                                      slice(int_center[1] - patch_radius, int_center[1] + patch_radius)]
                    storm_objects[-1].append(STObject(data[t][obj_slice_buff],
                                                      np.where(label_grid[t][obj_slice_buff] == o + 1, 1, 0),
                                                      x_grid[obj_slice_buff],
                                                      y_grid[obj_slice_buff],
                                                      ij_grid[0][obj_slice_buff],
                                                      ij_grid[1][obj_slice_buff],
                                                      time,
                                                      time,
                                                      dx=dx,
                                                      step=dt))
                    if t > 0:
                        dims = storm_objects[-1][-1].timesteps[0].shape
                        storm_objects[-1][-1].estimate_motion(time, data[t - 1], dims[1], dims[0])
    else:
        ij_grid = np.indices(label_grid.shape)
        storm_objects.append([])
        centers = list(center_of_mass(data, labels=label_grid, index=np.arange(1, label_grid.max() + 1)))

        if len(centers) > 0:
            for o, center in enumerate(centers):
                int_center = np.round(center).astype(int)
                obj_slice_buff = (slice(int_center[0] - patch_radius, int_center[0] + patch_radius),
                                  slice(int_center[1] - patch_radius, int_center[1] + patch_radius))
                storm_objects[-1].append(STObject(data[obj_slice_buff],
                                                  np.where(label_grid[obj_slice_buff] == o + 1, 1, 0),
                                                  x_grid[obj_slice_buff],
                                                  y_grid[obj_slice_buff],
                                                  ij_grid[0][obj_slice_buff],
                                                  ij_grid[1][obj_slice_buff],
                                                  times[0],
                                                  times[0],
                                                  dx=dx,
                                                  step=dt))
    return storm_objects


def track_storms(storm_objects, times, distance_components, distance_maxima, distance_weights, tracked_objects=None):
    """
    Given the output of extract_storm_objects, this method tracks storms through time and merges individual
    STObjects into a set of tracks.

    Args:
        storm_objects: list of list of STObjects that have not been tracked.
        times: List of times associated with each set of STObjects
        distance_components: list of function objects that make up components of distance function
        distance_maxima: array of maximum values for each distance for normalization purposes
        distance_weights: weight given to each component of the distance function. Should add to 1.
        tracked_objects: List of STObjects that have already been tracked.
    Returns:
        tracked_objects:
    """
    obj_matcher = ObjectMatcher(distance_components, distance_weights, distance_maxima)
    if tracked_objects is None:
        tracked_objects = []
    for t, time in enumerate(times):
        past_time_objects = []
        for obj in tracked_objects:
            if obj.end_time == time - obj.step:
                past_time_objects.append(obj)
        if len(past_time_objects) == 0:
            tracked_objects.extend(storm_objects[t])
        elif len(past_time_objects) > 0 and len(storm_objects[t]) > 0:
            assignments = obj_matcher.match_objects(past_time_objects, storm_objects[t], times[t-1], times[t])
            unpaired = list(range(len(storm_objects[t])))
            for pair in assignments:
                past_time_objects[pair[0]].extend(storm_objects[t][pair[1]])
                unpaired.remove(pair[1])
            if len(unpaired) > 0:
                for up in unpaired:
                    tracked_objects.append(storm_objects[t][up])
    return tracked_objects



__author__ = 'djgagne'

import unittest
from hagelslag.evaluation.ProbabilityMetrics import DistributedReliability, DistributedROC, DistributedCRPS
import numpy as np


class TestProbabilityMetrics(unittest.TestCase):
    def setUp(self):
        self.num_forecasts = 1000
        self.forecasts = dict(perfect=np.concatenate((np.ones(self.num_forecasts// 2), np.zeros(self.num_forecasts// 2))),
                              random=np.random.random(self.num_forecasts))
        self.observations= dict(perfect=self.forecasts['perfect'],
                                random=self.forecasts['perfect'])
        self.thresholds = np.arange(0, 1.2, 0.1)
        self.obs_threshold = 0.5
        return

    def test_reliability(self):
        perfect_rel = DistributedReliability(self.thresholds, self.obs_threshold)
        perfect_rel.update(self.forecasts["perfect"], self.observations["perfect"])
        random_rel = DistributedReliability(self.thresholds, self.obs_threshold)
        random_rel.update(self.forecasts["random"], self.observations["random"])
        perfect_components = perfect_rel.brier_score_components()
        self.assertEqual(perfect_rel.frequencies["Total_Freq"].sum(), self.num_forecasts,
                         msg="Total Frequency does not match number of forecasts.")
        self.assertEqual(perfect_rel.frequencies["Positive_Freq"].sum(), self.num_forecasts / 2,
                         msg="Positive Frequency does not match number of positive forecasts.")
        self.assertEqual(perfect_components[1], perfect_components[2], "Resolution does not equal uncertainty.")
        self.assertEqual(perfect_rel.brier_score(), 0,
                         msg="Perfect Brier score is {0:0.3f}".format(perfect_rel.brier_score()))
        self.assertGreater(random_rel.brier_score(), perfect_rel.brier_score(),
                           msg="Perfect (BS={0:0.3f}) has worse score than random (BS={1:0.3f})".format(
                               perfect_rel.brier_score(), random_rel.brier_score()))
        perfect_rel_copy = DistributedReliability(input_str=str(perfect_rel))
        self.assertEqual(perfect_rel.brier_score(), perfect_rel_copy.brier_score(),
                         msg="Brier Score of copy {0} does not match original {1}".format(perfect_rel.brier_score(),
                                                                                          perfect_rel_copy.brier_score()
                                                                                          ))
        pbss = perfect_rel.brier_skill_score()
        cpbss = perfect_rel_copy.brier_skill_score()
        self.assertEqual(pbss, cpbss,
                         msg="BSS of copy {0} does not match original {1}".format(pbss, cpbss))
        self.assertLessEqual(perfect_rel.frequencies["Positive_Freq"].sum(),
                             perfect_rel.frequencies["Total_Freq"].sum(),
                             msg="There are more perfect positives than total events")
        self.assertLessEqual(random_rel.frequencies["Positive_Freq"].sum(),
                             random_rel.frequencies["Total_Freq"].sum(),
                             msg="There are more random positives than total events")
        perfect_sum = perfect_rel + perfect_rel
        mixed_sum = perfect_rel + random_rel
        self.assertEqual(perfect_rel.brier_score(), perfect_sum.brier_score(),
                         msg="Summed perfect brier score not equal to perfect brier score")
        self.assertLess(perfect_sum.brier_score(), mixed_sum.brier_score(),
                           msg="Perfect brier score greater than mixed brier score")

    def test_roc(self):
        perfect_roc = DistributedROC(self.thresholds, self.obs_threshold)
        perfect_roc.update(self.forecasts["perfect"], self.observations["perfect"])
        perfect_auc = perfect_roc.auc()
        random_roc = DistributedROC(self.thresholds, self.obs_threshold)
        random_roc.update(self.forecasts["random"], self.observations["random"])
        random_auc = random_roc.auc()
        self.assertEqual(perfect_auc, 1, msg="Perfect AUC not 1, is actually {0:0.2f}".format(perfect_auc))
        self.assertLessEqual(np.abs(random_auc - 0.5), 0.1,
                             msg="Random AUC not 0.5, actually {0:0.3f}".format(random_auc))
        self.assertGreater(perfect_auc, random_auc, msg="Perfect AUC is not greater than random.")

    def test_crps(self):
        thresholds = np.arange(100)
        obs = np.zeros((1000, 100))
        for o in range(obs.shape[1]):
            ob_ix = np.reshape(np.arange(0, 1000, 100) + o, (10, 1))
            obs[ob_ix, thresholds[o:].reshape(1, 100 - o)] = 1
        perfect_crps = DistributedCRPS(thresholds=thresholds)
        perfect_crps.update(obs, obs)
        self.assertEqual(perfect_crps.crps(), 0, "CRPS for perfect forecast is not 0")
        self.assertGreater(perfect_crps.crps_climo(), 0, "Climo CRPS is greater than 0")
        self.assertLess(perfect_crps.crps_climo(), 1, "Climo CRPS is less than 1")
        self.assertEqual(perfect_crps.crpss(), 1,
                         "CRPSS for perfect forecast is not 1, is {0}".format(perfect_crps.crpss()))
        crps_copy = DistributedCRPS(input_str=str(perfect_crps))
        self.assertEqual(crps_copy.crps(), 0, "CRPS copy is not 0")
        self.assertEqual(crps_copy.crpss(), 1, "CRPSS copy is not 1")
        crps_sum = perfect_crps + perfect_crps
        self.assertEqual(crps_sum.crps(), 0, "CRPS sum is not 0")
        self.assertEqual(crps_sum.crpss(), 1, "CRPSS sum is not 1")



from netCDF4 import Dataset
from os.path import isdir, exists, join
import numpy as np
from glob import glob
import arrow


class WRFModelGrid(object):
    """
    Load data from raw WRF model output files.
    """
    def __init__(self, forecast_date, variable, domain, path, nocolons=False):
        self.forecast_date = arrow.get(forecast_date)
        self.variable = variable
        self.domain = domain
        self.path = path
        self.wrf_filename = "wrfout_d{0:02d}_{1}".format(self.domain, self.forecast_date.format("YYYY-MM-DD_HH:mm:ss"))
        if nocolons:
            self.wrf_filename = self.wrf_filename.replace(":", "_")
        self.patch_files = False
        if exists(join(self.path, self.wrf_filename)):
            if isdir(join(self.path, self.wrf_filename)):
                self.patch_files = True
        else:
            raise IOError(join(self.path, self.wrf_filename) + " not found.")
        return

    def get_global_attributes(self):
        attributes = {}
        if self.patch_files:
            patch_files = sorted(glob(join(self.path, "{0}/{0}_".format(self.wrf_filename) + "[0-9]" * 5)))
            wrf_data = Dataset(patch_files[0])
        else:
            wrf_data = Dataset(join(self.path, self.wrf_filename))
        for attr in wrf_data.ncattrs():
            attributes[attr] = getattr(wrf_data, attr)
        wrf_data.close()
        return attributes

    def load_time_var(self, time_var="XTIME"):
        var_attrs = dict()
        var_val = 0
        if self.patch_files:
            patch_file_list = sorted(glob(join(self.path, "{0}/{0}_".format(self.wrf_filename) + "[0-9]" * 5)))
            patch_zero = Dataset(patch_file_list[0])
            var_list = patch_zero.variables.keys()
            if time_var in var_list:
                for attr in patch_zero.variables[time_var].ncattrs():
                    var_attrs[attr] = getattr(patch_zero.variables[time_var], attr)
                var_val = patch_zero.variables[time_var][:]        
            patch_zero.close()
        else:
            wrf_data = Dataset(join(self.path, self.wrf_filename))
            var_list = wrf_data.variables.keys()
            if time_var in var_list:
                for attr in wrf_data.variables[time_var].ncattrs():
                    var_attrs[attr] = getattr(wrf_data.variables[time_var], attr)
                var_val = wrf_data.variables[time_var][:]   
            wrf_data.close()
        return var_val, var_attrs 

    def load_full_grid(self):
        var_data = None
        var_attrs = dict()
        if self.patch_files:
            patch_file_list = sorted(glob(join(self.path, "{0}/{0}_".format(self.wrf_filename) + "[0-9]" * 5)))
            patch_zero = Dataset(patch_file_list[0])
            var_list = patch_zero.variables.keys()
            if self.variable in var_list:
                dimension_names = patch_zero.variables[self.variable].dimensions
                is_stag = np.array(["stag" in x for x in patch_zero.variables[self.variable].dimensions])
                for attr in patch_zero.variables[self.variable].ncattrs():
                    if attr == "coordinates":
                        var_attrs[attr] = getattr(patch_zero.variables[self.variable], attr)
                        if "_U" in var_attrs[attr]:
                            var_attrs[attr] = var_attrs[attr].replace("_U", "")
                        elif "_V" in var_attrs[attr]:
                            var_attrs[attr] = var_attrs[attr].replace("_V", "")
                    else:
                        var_attrs[attr] = getattr(patch_zero.variables[self.variable], attr)
                if len(dimension_names) == 4:
                    grid_dim = (1,
                                getattr(patch_zero, "BOTTOM-TOP_GRID_DIMENSION") - 1,
                                getattr(patch_zero, "SOUTH-NORTH_GRID_DIMENSION") - 1,
                                getattr(patch_zero, "WEST-EAST_GRID_DIMENSION") - 1)
                else:
                    grid_dim = (1,
                                getattr(patch_zero, "SOUTH-NORTH_GRID_DIMENSION") - 1,
                                getattr(patch_zero, "WEST-EAST_GRID_DIMENSION") - 1)
                patch_zero.close()
                if np.any(is_stag):
                    grid_dim_arr = np.array(grid_dim)
                    grid_dim_arr[is_stag] += 1
                    var_data = np.zeros(grid_dim_arr, dtype=np.float32)
                    stag_dim = np.where(is_stag)[0][0]
                else:
                    var_data = np.zeros(grid_dim, dtype=np.float32)
                    stag_dim = None
                for p, patch_file in enumerate(patch_file_list):
                    wrf_patch_file = Dataset(patch_file)
                    sub_data = wrf_patch_file.variables[self.variable][:]
                    if stag_dim is not None:
                        if len(sub_data.shape) - stag_dim == 1:
                            we_slice = slice(getattr(wrf_patch_file, "WEST-EAST_PATCH_START_STAG") - 1,
                                             getattr(wrf_patch_file, "WEST-EAST_PATCH_END_STAG"))
                            sn_slice = slice(getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_START_UNSTAG") - 1,
                                             getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_END_UNSTAG"))
                        elif len(sub_data.shape) - stag_dim == 2:
                            we_slice = slice(getattr(wrf_patch_file, "WEST-EAST_PATCH_START_UNSTAG") - 1,
                                             getattr(wrf_patch_file, "WEST-EAST_PATCH_END_UNSTAG"))
                            sn_slice = slice(getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_START_STAG") - 1,
                                             getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_END_STAG"))
                        else:
                            we_slice = slice(getattr(wrf_patch_file, "WEST-EAST_PATCH_START_UNSTAG") - 1,
                                             getattr(wrf_patch_file, "WEST-EAST_PATCH_END_UNSTAG"))
                            sn_slice = slice(getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_START_UNSTAG") - 1,
                                             getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_END_UNSTAG"))
                    else:
                        we_slice = slice(getattr(wrf_patch_file, "WEST-EAST_PATCH_START_UNSTAG") - 1,
                                         getattr(wrf_patch_file, "WEST-EAST_PATCH_END_UNSTAG"))
                        sn_slice = slice(getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_START_UNSTAG") - 1,
                                         getattr(wrf_patch_file, "SOUTH-NORTH_PATCH_END_UNSTAG"))
                    if len(var_data.shape) == 3:
                        var_data[:, sn_slice, we_slice] = sub_data
                    else:
                        var_data[:, :, sn_slice, we_slice] = sub_data
                    wrf_patch_file.close()
                if stag_dim is not None:
                    if stag_dim == 1:
                        var_data = 0.5 * (var_data[:, :-1] + var_data[:, 1:])
                    elif stag_dim == 2:
                        var_data = 0.5 * (var_data[:, :, :-1] + var_data[:, :, 1:])
                    else:
                        var_data = 0.5 * (var_data[:, :, :, :-1] + var_data[:, :, :, 1:])
            else:
                patch_zero.close()

        else:
            wrf_data = Dataset(join(self.path, self.wrf_filename))
            is_stag = np.array(["stag" in x for x in wrf_data.variables[self.variable].dimensions])
            if self.variable in wrf_data.variables.keys():
                for attr in wrf_data.variables[self.variable].ncattrs():
                    if attr == "coordinates":
                        var_attrs[attr] = getattr(wrf_data.variables[self.variable], attr)
                        if "_U" in var_attrs[attr]:
                            var_attrs[attr] = var_attrs[attr].replace("_U", "")
                        elif "_V" in var_attrs[attr]:
                            var_attrs[attr] = var_attrs[attr].replace("_V", "")
                    else:
                        var_attrs[attr] = getattr(wrf_data.variables[self.variable], attr)

                var_data = wrf_data.variables[self.variable][:] 
                if np.any(is_stag):
                    stag_dim = np.where(is_stag)[0][0]
                    if stag_dim == 1:
                        var_data = 0.5 * (var_data[:, :-1] + var_data[:, 1:])
                    elif stag_dim == 2:
                        var_data = 0.5 * (var_data[:, :, :-1] + var_data[:, :, 1:])
                    else:
                        var_data = 0.5 * (var_data[:, :, :, :-1] + var_data[:, :, :, 1:])
            wrf_data.close()

        return var_data, var_attrs



#!/home/tmp/aburke/miniconda3/bin/python -u
from hagelslag.util.make_proj_grids import read_ncar_map_file
from netCDF4 import Dataset
import pandas as pd
import numpy as np
import os
from netCDF4 import Dataset
from os.path import exists

class SectorProcessor(object):
    def __init__(self,mapfile,
                ensemble_name,member,
                run_date,date_format):

        self.mapfile = mapfile
        self.ensemble_name = ensemble_name
        self.member = member
        self.run_date = run_date
        self.date_format = date_format
        self.inds = None 
        proj_dict, grid_dict = read_ncar_map_file(self.mapfile) 
        
        self.ne_lat, self.sw_lat = grid_dict["ne_lat"],grid_dict["sw_lat"]
        self.ne_lon, self.sw_lon = grid_dict["ne_lon"],grid_dict["sw_lon"]

        return
    
    def output_sector_csv(self,csv_path,file_dict_key,out_path):
        """
        Segment forecast tracks to only output data contined within a 
        region in the CONUS, as defined by the mapfile.

        Args:
            csv_path(str): Path to the full CONUS csv file.
            file_dict_key(str): Dictionary key for the csv files, 
                currently either 'track_step' or 'track_total'
            out_path (str): Path to output new segmented csv files.
        Returns:
            Segmented forecast tracks in a csv file.
        """
        csv_file = csv_path + "{0}_{1}_{2}_{3}.csv".format(
                                                        file_dict_key,
                                                        self.ensemble_name,
                                                        self.member,
                                                        self.run_date.strftime(self.date_format))
        if exists(csv_file):
            csv_data = pd.read_csv(csv_file)
            
            if self.inds is None:
                lon_obj = csv_data.loc[:,"Centroid_Lon"]
                lat_obj = csv_data.loc[:,"Centroid_Lat"]
            
                self.inds = np.where((self.ne_lat>=lat_obj)&(self.sw_lat<=lat_obj)\
                        &(self.ne_lon>=lon_obj)&(self.sw_lon<=lon_obj))[0]
            
            if np.shape(self.inds)[0] > 0:
                csv_data = csv_data.reindex(np.array(self.inds)) 
                sector_csv_filename = out_path + "{0}_{1}_{2}_{3}.csv".format(
                                                        file_dict_key,
                                                        self.ensemble_name,
                                                        self.member,
                                                        self.run_date.strftime(self.date_format))
                print("Output sector csv file " + sector_csv_filename)
                csv_data.to_csv(sector_csv_filename,
                        na_rep="nan",
                        float_format="%0.5f",
                        index=False)
                os.chmod(sector_csv_filename, 0o666)
            else:
                print('No {0} {1} sector data found'.format(self.member,
                                self.run_date.strftime("%Y%m%d")))
            
        else:
            print('No {0} {1} csv file found'.format(self.member,
                                self.run_date.strftime("%Y%m%d")))
        return 


    def load_netcdf_data(self,netcdf_path,patch_radius):
        nc_file = netcdf_path+ "{0}_{1}_{2}_model_patches.nc".format(
                                                            self.ensemble_name,
                                                            self.run_date.strftime(self.date_format),
                                                            self.member)
        if exists(nc_file):
            nc_data = Dataset(nc_file)
            
            lon_obj = nc_data.variables['centroid_lon'][:]
            lat_obj = nc_data.variables['centroid_lat'][:]
            
            inds = np.where((self.ne_lat>=lat_obj)&(self.sw_lat<=lat_obj)\
                        &(self.ne_lon>=lon_obj)&(self.sw_lon<=lon_obj))[0]
            if np.shape(inds)[0] > 0:
                for var in nc_data.variables:
                    if np.shape(nc_data.variables[var])[0] > (2*patch_radius):
                        try:
                            nc_data.variables[var] =\
                            nc_data.variables[var][np.array(inds),:,:]
                        except:
                            nc_data.variables[var] =\
                                nc_data.variables[var][np.array(inds)]
            else:
                nc_data=None
        else:
            nc_data=None
        
        return nc_data
    
    def output_sector_netcdf(self,netcdf_path,out_path,patch_radius,config):
        """
        Segment patches of forecast tracks to only output data contined within a 
        region in the CONUS, as defined by the mapfile.

        Args:
            netcdf_path (str): Path to the full CONUS netcdf patch file.
            out_path (str): Path to output new segmented netcdf files.
            patch_radius (int): Size of the patch radius.
            config (dict): Dictonary containing information about data and
                            ML variables
        Returns:
            Segmented patch netcdf files.
        """
        
        nc_data = self.load_netcdf_data(netcdf_path,patch_radius)
    
        if nc_data is not None:
            out_filename = out_path + "{0}_{1}_{2}_model_patches.nc".format(
                                                            self.ensemble_name,
                                                            self.run_date.strftime(self.date_format),
                                                            self.member)
            out_file = Dataset(out_filename, "w")
            out_file.createDimension("p", np.shape(nc_data.variables['p'])[0])
            out_file.createDimension("row", np.shape(nc_data.variables['row'])[0])
            out_file.createDimension("col", np.shape(nc_data.variables['col'])[0])
            out_file.createVariable("p", "i4", ("p",))
            out_file.createVariable("row", "i4", ("row",))
            out_file.createVariable("col", "i4", ("col",))
            out_file.variables["p"][:] = nc_data.variables['p'][:]
            out_file.variables["row"][:] =  nc_data.variables['row'][:]
            out_file.variables["col"][:] =  nc_data.variables['col'][:]
            out_file.Conventions = "CF-1.6"
            out_file.title = "{0} Storm Patches for run {1} member {2}".format(self.ensemble_name,
                                                                       self.run_date.strftime(self.date_format),
                                                                       self.member)
            out_file.object_variable = config.watershed_variable
            meta_variables = ["lon", "lat", "i", "j", "x", "y", "masks"]
            meta_units = ["degrees_east", "degrees_north", "", "", "m", "m", ""]
            center_vars = ["time", "centroid_lon", "centroid_lat", "centroid_i", "centroid_j", "track_id", "track_step"]
            center_units = ["hours since {0}".format(self.run_date.strftime("%Y-%m-%d %H:%M:%S")),
                    "degrees_east",
                    "degrees_north",
                    "",
                    "",
                    "",
                    ""]

            label_columns = ["Matched", "Max_Hail_Size", "Num_Matches", "Shape", "Location", "Scale"]
        
            for m, meta_variable in enumerate(meta_variables):
                if meta_variable in ["i", "j", "masks"]:
                    dtype = "i4"
                else:
                    dtype = "f4"
                m_var = out_file.createVariable(meta_variable, dtype, ("p", "row", "col"), complevel=1, zlib=True)
                m_var.long_name = meta_variable
                m_var.units = meta_units[m]
        
            for c, center_var in enumerate(center_vars):
                if center_var in ["time", "track_id", "track_step"]:
                    dtype = "i4"
                else:
                    dtype = "f4"
                c_var = out_file.createVariable(center_var, dtype, ("p",), zlib=True, complevel=1)
                c_var.long_name = center_var
                c_var.units =center_units[c]
        
            for storm_variable in config.storm_variables:
                s_var = out_file.createVariable(storm_variable + "_curr", "f4", ("p", "row", "col"), complevel=1, zlib=True)
                s_var.long_name = storm_variable
                s_var.units = ""
        
            for potential_variable in config.potential_variables:
                p_var = out_file.createVariable(potential_variable + "_prev", "f4", ("p", "row", "col"),
                                        complevel=1, zlib=True)
                p_var.long_name = potential_variable
                p_var.units = ""
        
            if config.train:
                for label_column in label_columns:
                    if label_column in ["Matched", "Num_Matches"]:
                        dtype = "i4"
                    else:
                        dtype = "f4"
                    l_var = out_file.createVariable(label_column, dtype, ("p",), zlib=True, complevel=1)
                    l_var.long_name = label_column
                    l_var.units = ""
            
            out_file.variables["time"][:] = nc_data.variables['time'][:]
    
            for c_var in ["lon", "lat"]:
                out_file.variables["centroid_" + c_var][:] =  nc_data.variables['centroid_' + c_var][:]

            for c_var in ["i", "j"]:
                out_file.variables["centroid_" + c_var][:] =  nc_data.variables["centroid_" + c_var][:]
        
            out_file.variables["track_id"][:] =  nc_data.variables['track_id'][:]
            out_file.variables["track_step"][:] =  nc_data.variables['track_step'][:]
        
            for meta_var in meta_variables:
                if meta_var in ["lon", "lat"]:
                    out_file.variables[meta_var][:] = nc_data.variables[meta_var][:]
                else:
                    out_file.variables[meta_var][:] = nc_data.variables[meta_var][:]
        
            for storm_variable in config.storm_variables:
                out_file.variables[storm_variable + "_curr"][:] = nc_data.variables[storm_variable + '_curr'][:]
        
            for p_variable in config.potential_variables:
                out_file.variables[p_variable + "_prev"][:] = nc_data.variables[p_variable + '_prev'][:]
        
            if config.train:
                for label_column in label_columns:
                    try:
                        out_file.variables[label_column][:] = nc_data.variables[label_column][:]
                    except:
                        out_file.variables[label_column][:] = 0

            out_file.close()
        
            print("Output sector nc file " + out_filename)
        else:
            print('No {0} {1} netcdf file/sector data found'.format(self.member,
                                self.run_date.strftime("%Y%m%d")))

        return

if __name__ == "__main__":
    main()



import numpy as np


class ContingencyTable(object):
    """
    Initializes a binary contingency table and generates many skill scores.

    Args:
        a: true positives
        b: false positives
        c: false negatives
        d: true negatives

    Attributes:
        table (numpy.ndarray): contingency table
        N: total number of items in table

    """
    def __init__(self, a, b, c, d):
        self.table = np.array([[a, b], [c, d]], dtype=float)
        self.N = self.table.sum()

    def update(self, a, b, c, d):
        """
        Update contingency table with new values without creating a new object.
        """
        self.table.ravel()[:] = [a, b, c, d]
        self.N = self.table.sum()

    def __add__(self, other):
        """
        Add two contingency tables together and return a combined one.

        Args:
            other: Another contingency table

        Returns:
            Sum of contingency tables
        """
        sum_ct = ContingencyTable(*(self.table + other.table).tolist())
        return sum_ct

    def pod(self):
        """
        Probability of Detection (POD) or Hit Rate.
        Formula:  a/(a+c)
        """
        return self.table[0, 0] / (self.table[0, 0] + self.table[1, 0])

    def foh(self):
        """
        Frequency of Hits (FOH) or Success Ratio.
        Formula:  a/(a+b)
        """
        return self.table[0, 0] / (self.table[0, 0] + self.table[0, 1])

    def far(self):
        """
        False Alarm Ratio (FAR).
        Formula:  b/(a+b)
        """
        return self.table[0, 1] / (self.table[0, 0] + self.table[0, 1])

    def pofd(self):
        """
        Probability of False Detection (POFD).
        b/(b+d)
        """
        return self.table[0, 1] / (self.table[0, 1] + self.table[1, 1])

    def fom(self):
        """
        Frequency of Misses (FOM).
        Formula:  c/(a+c)."""
        return self.table[1, 0] / (self.table[0, 0] + self.table[1, 0])

    def dfr(self):
        """Returns Detection Failure Ratio (DFR).
           Formula:  c/(c+d)"""
        return self.table[1, 0] / (self.table[1, 0] + self.table[1, 1])

    def pon(self):
        """Returns Probability of Null (PON).
           Formula:  d/(b+d)"""
        return self.table[1, 1] / (self.table[0, 1] + self.table[1, 1])

    def focn(self):
        """Returns Frequency of Correct Null (FOCN).
           Formula:  d/(c+d)"""
        return self.table[1, 1] / (self.table[1, 0] + self.table[1, 1])

    def bias(self):
        """
        Frequency Bias.
        Formula:  (a+b)/(a+c)"""
        return (self.table[0, 0] + self.table[0, 1]) / (self.table[0, 0] + self.table[1, 0])

    def accuracy(self):
        """Finley's measure, fraction correct, accuracy (a+d)/N"""
        return (self.table[0, 0] + self.table[1, 1]) / self.N

    def csi(self):
        """Gilbert's Score or Threat Score or Critical Success Index a/(a+b+c)"""
        return self.table[0, 0] / (self.table[0, 0] + self.table[0, 1] + self.table[1, 0])

    def ets(self):
        """Equitable Threat Score, Gilbert Skill Score, v, (a - R)/(a + b + c - R), R=(a+b)(a+c)/N"""
        r = (self.table[0, 0] + self.table[0, 1]) * (self.table[0, 0] + self.table[1, 0]) / self.N
        return (self.table[0, 0] - r) / (self.table[0, 0] + self.table[0, 1] + self.table[1, 0] - r)

    def hss(self):
        """Doolittle (Heidke) Skill Score.  2(ad-bc)/((a+b)(b+d) + (a+c)(c+d))"""
        return 2 * (self.table[0, 0] * self.table[1, 1] - self.table[0, 1] * self.table[1, 0]) / (
            (self.table[0, 0] + self.table[0, 1]) * (self.table[0, 1] + self.table[1, 1]) +
            (self.table[0, 0] + self.table[1, 0]) * (self.table[1, 0] + self.table[1, 1]))

    def pss(self):
        """Peirce (Hansen-Kuipers, True) Skill Score (ad - bc)/((a+c)(b+d))"""
        return (self.table[0, 0] * self.table[1, 1] - self.table[0, 1] * self.table[1, 0]) / \
               ((self.table[0, 0] + self.table[1, 0]) * (self.table[0, 1] + self.table[1, 1]))

    def css(self):
        """Clayton Skill Score (ad - bc)/((a+b)(c+d))"""
        return (self.table[0, 0] * self.table[1, 1] - self.table[0, 1] * self.table[1, 0]) / \
               ((self.table[0, 0] + self.table[0, 1]) * (self.table[1, 0] + self.table[1, 1]))

    def __str__(self):
        table_string = '\tEvent\n\tYes\tNo\nYes\t%d\t%d\nNo\t%d\t%d\n' % (
            self.table[0, 0], self.table[0, 1], self.table[1, 0], self.table[1, 1])
        return table_string


if __name__ == "__main__":
    table = ContingencyTable(1, 2, 3, 4)
    print(table)
    print(getattr(table, "pss")())



from netCDF4 import Dataset
import numpy as np
from pandas import DatetimeIndex
from os.path import exists


class ModelGrid(object):
    """
    Base class for reading 2D model output grids from netCDF files.

    Given a list of file names, loads the values of a single variable from a model run. Supports model output in
    netCDF format.

    Attributes:
        filenames (list of str): List of netCDF files containing model output
        run_date (ISO date string or datetime.datetime object): Date of the initialization time of the model run.
        start_date (ISO date string or datetime.datetime object): Date of the first timestep extracted.
        end_date (ISO date string or datetime.datetime object): Date of the last timestep extracted.
        freqency (str): spacing between model time steps.
        valid_dates: DatetimeIndex of all model timesteps
        forecast_hours: array of all hours in the forecast
        file_objects (list): List of the file objects for each model time step
    """
    def __init__(self, 
                 filenames, 
                 run_date, 
                 start_date, 
                 end_date,
                 variable,
                 frequency="1H"):
        self.filenames = filenames
        self.variable = variable
        self.run_date = np.datetime64(run_date)
        self.start_date = np.datetime64(start_date)
        self.end_date = np.datetime64(end_date)
        self.frequency = frequency
        self.valid_dates = DatetimeIndex(start=self.start_date,
                                         end=self.end_date,
                                         freq=self.frequency)
        self.forecast_hours = (self.valid_dates.values - self.run_date).astype("timedelta64[h]").astype(int)
        self.file_objects = []
        self.__enter__()

    def __enter__(self):
        """
        Open each file for reading.

        """
        for filename in self.filenames:
            if exists(filename):
                self.file_objects.append(Dataset(filename))
            else:
                self.file_objects.append(None)

    def load_data_old(self):
        """
        Loads time series of 2D data grids from each opened file. The code 
        handles loading a full time series from one file or individual time steps
        from multiple files. Missing files are supported.
        """
        units = ""
        if len(self.file_objects) == 1 and self.file_objects[0] is not None:
            data = self.file_objects[0].variables[self.variable][self.forecast_hours]
            if hasattr(self.file_objects[0].variables[self.variable], "units"):
                units = self.file_objects[0].variables[self.variable].units
        elif len(self.file_objects) > 1:
            grid_shape = [len(self.file_objects), 1, 1]
            for file_object in self.file_objects:
                if file_object is not None:
                    if self.variable in file_object.variables.keys():
                        grid_shape = file_object.variables[self.variable].shape
                    elif self.variable.ljust(6, "_") in file_object.variables.keys():
                        grid_shape = file_object.variables[self.variable.ljust(6, "_")].shape

                    else:
                        print("{0} not found".format(self.variable))
                        raise KeyError
                    break
            data = np.zeros((len(self.file_objects), grid_shape[1], grid_shape[2]))
            for f, file_object in enumerate(self.file_objects):
                if file_object is not None:
                    if self.variable in file_object.variables.keys():
                        var_name = self.variable
                    elif self.variable.ljust(6, "_") in file_object.variables.keys():
                        var_name = self.variable.ljust(6, "_")
                    else:
                        print("{0} not found".format(self.variable))
                        raise KeyError
                    data[f] = file_object.variables[var_name][0]
                    if units == "" and hasattr(file_object.variables[var_name], "units"):
                        units = file_object.variables[var_name].units
        else:
            data = None
        return data, units

    def load_data(self):
        """
        Load data from netCDF file objects or list of netCDF file objects. Handles special variable name formats.

        Returns:
            Array of data loaded from files in (time, y, x) dimensions, Units
        """
        units = ""
        if self.file_objects[0] is None:
            raise IOError()
        var_name, z_index = self.format_var_name(self.variable, list(self.file_objects[0].variables.keys()))
        ntimes = 0
        if 'time' in self.file_objects[0].variables[var_name].dimensions:
            ntimes = len(self.file_objects[0].dimensions['time'])

        if ntimes > 1:
            if z_index is None:
                data = self.file_objects[0].variables[var_name][self.forecast_hours].astype(np.float32)
            else:
                data = self.file_objects[0].variables[var_name][self.forecast_hours, z_index].astype(np.float32)
        else:
            y_dim, x_dim = self.file_objects[0].variables[var_name].shape[-2:]
            data = np.zeros((len(self.valid_dates), y_dim, x_dim), dtype=np.float32)
            for f, file_object in enumerate(self.file_objects):
                if file_object is not None:
                    if z_index is None:
                        data[f] = file_object.variables[var_name][0]
                    else:
                        data[f] = file_object.variables[var_name][0, z_index]
        if hasattr(self.file_objects[0].variables[var_name], "units"):
            units = self.file_objects[0].variables[var_name].units
        return data, units

    @staticmethod
    def format_var_name(variable, var_list):
        """
        Searches var list for variable name, checks other variable name format options.

        Args:
            variable (str): Variable being loaded
            var_list (list): List of variables in file.

        Returns:
            Name of variable in file containing relevant data, and index of variable z-level if multiple variables
            contained in same array in file.
        """
        z_index = None
        if variable in var_list:
            var_name = variable
        elif variable.ljust(6, "_") in var_list:
            var_name = variable.ljust(6, "_")
        elif any([variable in v_sub.split("_") for v_sub in var_list]):
            var_name = var_list[[variable in v_sub.split("_") for v_sub in var_list].index(True)]
            z_index = var_name.split("_").index(variable)
        else:
            raise KeyError("{0} not found in {1}".format(variable, var_list))
        return var_name, z_index

    def __exit__(self):
        """
        Close links to all open file objects and delete the objects.
        """
        for file_object in self.file_objects:
            file_object.close()
        del self.file_objects[:]

    def close(self):
        """
        Close links to all open file objects and delete the objects.
        """
        self.__exit__()




#!/usr/bin/env python
import os
from pyproj import Proj
from scipy.spatial import cKDTree

import pygrib
import numpy as np
from os.path import exists
import pandas as pd
from pandas import DatetimeIndex


class ModelGrid(object):
    """
    Base class for reading 2D model output grids from grib2 files.
    Given a list of file names, loads the values of a single variable from a model run. Supports model output in
    grib2 format
    Attributes:
            filenames (list of str): List of grib2 files containing model output
            run_date (ISO date string or datetime.datetime object): Date of the initialization time of the model run.
            start_date (ISO date string or datetime.datetime object): Date of the first timestep extracted.
            end_date (ISO date string or datetime.datetime object): Date of the last timestep extracted.
            freqency (str): spacing between model time steps.
            valid_dates(dattime.datetime): DatetimeIndex of all model timesteps
            forecast_hours(array): array of all hours in the forecast
            file_objects (list): List of the file objects for each model time step
    """

    def __init__(self, filenames,
                 run_date,
                 start_date,
                 end_date,
                 variable,
                 member,
                 mapping_data,
                 sector_ind_path,
                 frequency="1H"):

        self.filenames = filenames
        self.variable = variable
        self.run_date = np.datetime64(run_date)
        self.start_date = np.datetime64(start_date)
        self.end_date = np.datetime64(end_date)
        self.frequency = frequency
        self.valid_dates = DatetimeIndex(start=self.start_date,
                                         end=self.end_date,
                                         freq=self.frequency)
        self.forecast_hours = (self.valid_dates.values - self.run_date).astype("timedelta64[h]").astype(int)
        self.file_objects = []
        self.member = member
        self.mapping_data = mapping_data
        self.sector_ind_path = sector_ind_path
        self.__enter__()
        self.data = None
        self.lat = None
        self.lon = None
        self.unknown_names = {3: "LCDC", 4: "MCDC", 5: "HCDC", 197: "RETOP", 198: "MAXREF", 199: "MXUPHL",
                              200: "MNUPHL", 220: "MAXUVV", 221: "MAXDVV", 222: "MAXUW", 223: "MAXVW"}
        self.unknown_units = {3: "%", 4: "%", 5: "%", 197: "m", 198: "dB", 199: "m**2 s**-2", 200: "m**2 s**-2",
                              220: "m s**-1", 221: "m s**-1", 222: "m s**-1", 223: "m s**-1"}

    def __enter__(self):
        """
        Open each file for reading.
        """
        for filename in self.filenames:
            if exists(filename):
                self.file_objects.append(filename)

    def format_grib_name(self, selected_variable):
        """
        Assigns name to grib2 message number with name 'unknown'. Names based on NOAA grib2 abbreviations.
        Args:
            selected_variable(str): name of selected variable for loading
        Names:
            3: LCDC: Low Cloud Cover
            4: MCDC: Medium Cloud Cover
            5: HCDC: High Cloud Cover
            197: RETOP: Echo Top
            198: MAXREF: Hourly Maximum of Simulated Reflectivity at 1 km AGL
            199: MXUPHL: Hourly Maximum of Updraft Helicity over Layer 2km to 5 km AGL, and 0km to 3km AGL
                    examples:' MXUPHL_5000' or 'MXUPHL_3000'
            200: MNUPHL: Hourly Minimum of Updraft Helicity at same levels of MXUPHL
                     examples:' MNUPHL_5000' or 'MNUPHL_3000'
            220: MAXUVV: Hourly Maximum of Upward Vertical Velocity in the lowest 400hPa
            221: MAXDVV: Hourly Maximum of Downward Vertical Velocity in the lowest 400hPa
            222: MAXUW: U Component of Hourly Maximum 10m Wind Speed
            223: MAXVW: V Component of Hourly Maximum 10m Wind Speed
        Returns:
            Given an uknown string name of a variable, returns the grib2 message Id
            and units of the variable, based on the self.unknown_name and
            self.unknown_units dictonaries above. Allows access of
            data values of unknown variable name, given the ID.
        """
        names = self.unknown_names
        units = self.unknown_units
        for key, value in names.items():
            if selected_variable == value:
                Id = key
                u = units[key]
        return Id, u

    def load_data(self):
        """
            Loads data from grib2 file objects or list of grib2 file objects. Handles specific grib2 variable names
            and grib2 message numbers.
            Returns:
                    Array of data loaded from files in (time, y, x) dimensions, Units
        """
        file_objects = self.file_objects
        var = self.variable
        valid_date = self.valid_dates
        data = self.data
        unknown_names = self.unknown_names
        unknown_units = self.unknown_units
        member = self.member
        lat = self.lat
        lon = self.lon
      
        if self.sector_ind_path:
            inds_file = pd.read_csv(self.sector_ind_path+'sector_data_indices.csv') 
            inds = inds_file.loc[:,'indices']  
        out_x = self.mapping_data["x"]
        
        if not file_objects:
            print()
            print("No {0} model runs on {1}".format(member,self.run_date))
            print()
            units = None
            return self.data, units

    
        for f, file in enumerate(file_objects):
            grib = pygrib.open(file)
            if type(var) is int:
                data_values = grib[var].values
                #lat, lon = grib[var].latlons()
                #proj = Proj(grib[var].projparams)
                if grib[var].units == 'unknown':
                    Id = grib[var].parameterNumber
                    units = self.unknown_units[Id] 
                else:
                    units = grib[var].units
            elif type(var) is str:
                if '_' in var:
                    variable = var.split('_')[0]
                    level = int(var.split('_')[1])
                    if variable in unknown_names.values():
                        Id, units = self.format_grib_name(variable)
                        data_values = grib.select(parameterNumber=Id, level=level)[0].values
                        #lat, lon =  grib.select(parameterNumber=Id, level=level)[0].latlons()
                        #proj = Proj(grib.select(parameterNumber=Id, level=level)[0].projparams)

                    else:
                        data_values = grib.select(name=variable, level=level)[0].values
                        units = grib.select(name=variable, level=level)[0].units
                        #lat, lon  = grib.select(name=variable, level=level)[0].latlons()
                        #proj = Proj(grib.select(name=variable, level=level)[0].projparams)
                else:
                    if var in unknown_names.values():
                        Id, units = self.format_grib_name(var)
                        data_values = grib.select(parameterNumber=Id)[0].values
                        #lat, lon = grib.select(parameterNumber=Id)[0].latlons() 
                        #proj = Proj(grib.select(parameterNumber=Id)[0].projparams)

                    elif len(grib.select(name=var)) > 1:
                        raise NameError("Multiple '{0}' records found. Rename with level:'{0}_level'".format(var))

                    else:
                        data_values = grib.select(name=var)[0].values
                        units = grib.select(name=var)[0].units
                        #lat, lon = grib.select(name=var)[0].latlons()
                        #proj = Proj(grib.select(name=var)[0].projparams)

            if data is None:
                data = np.empty((len(valid_date), out_x.shape[0], out_x.shape[1]), dtype=float)
                if self.sector_ind_path:
                    data[f] = data_values[:].flatten()[inds].reshape(out_x.shape)
                else:
                    data[f]=data_values[:]
            else:
                if self.sector_ind_path:
                    data[f] = data_values[:].flatten()[inds].reshape(out_x.shape)
                else:
                    data[f]=data_values[:]
        
        return data, units
        
    def __exit__(self):
        """
        Close links to all open file objects and delete the objects.
        """
        for file_object in self.file_objects:
            file_object.close()
        del self.file_objects[:]

    def close(self):
        """
        Close links to all open file objects and delete the objects.
        """
        self.__exit__()


class HREFv2ModelGrid(ModelGrid):
    """
    Extension of the ModelGrid class for interfacing with the HREFv2  ensemble.
    Args:
        member (str): Name of the ensemble member
        run_date (datetime.datetime object): Date of the initial step of the ensemble run
        variable(int or str): name of grib2 variable(str) or grib2 message number(int) being loaded
        start_date (datetime.datetime object): First time step extracted.
        end_date (datetime.datetime object): Last time step extracted.
        path (str): Path to model output files
        single_step (boolean (default=True): Whether variable information is stored with each time step in a separate
                file (True) or one file containing all timesteps (False).
    """

    def __init__(self, member, run_date, variable, start_date, 
                end_date, path,mapping_data,sector_ind_path,single_step=True):
        self.path = path
        self.member = member
        filenames = []
        self.forecast_hours = np.arange((start_date - run_date).total_seconds() / 3600,
                                        (end_date - run_date).total_seconds() / 3600 + 1, dtype=int)

        if 'nam' in self.member:
            if '00' in self.member:
                for forecast_hr in self.forecast_hours:
                    file = '{1}/{2}/{3}/nam.t00z.conusnest.camfld{0:02}.tm00.grib2'.format(forecast_hr,
                                                                                           self.path,
                                                                                           self.member,
                                                                                           run_date.strftime("%Y%m%d"))
                    filenames.append(file)
            elif '12' in self.member:
                for forecast_hr in self.forecast_hours:
                    file = '{1}/{2}/{3}/nam.t12z.conusnest.camfld{0:02}.tm00.grib2'.format(forecast_hr,
                                                                                           self.path,
                                                                                           self.member,
                                                                                           run_date.strftime("%Y%m%d"))
                    filenames.append(file)
        else:
            member_name = str(self.member.split("_")[0])
            if '00' in self.member:
                for forecast_hr in self.forecast_hours:
                    file = '{1}/{2}/{3}/hiresw_conus{4}_{3}00f0{0:02}.grib2'.format(forecast_hr,
                                                                                    self.path,
                                                                                    self.member,
                                                                                    run_date.strftime("%Y%m%d"),
                                                                                    member_name)
                    filenames.append(file)

            elif '12' in self.member:
                for forecast_hr in self.forecast_hours:
                    file = '{1}/{2}/{3}/hiresw_conus{4}_{3}12f0{0:02}.grib2'.format(forecast_hr,
                                                                                    self.path,
                                                                                    self.member,
                                                                                    run_date.strftime("%Y%m%d"),
                                                                                    member_name)
                    filenames.append(file)

        super(HREFv2ModelGrid, self).__init__(filenames, run_date, start_date, end_date, 
                                            variable, member,mapping_data,sector_ind_path)
        return



class Config(object):
    """
    Class that loads options from a config file and converts
    them into attributes.
    """
    def __init__(self, filename, required_attributes=()):
        config = None
        print(filename)
        config_file = open(filename)
        config_text = config_file.read()
        config_file.close()
        if "config" not in config_text:
            raise ValueError
        d = {}
        exec(config_text, d)
        config = d["config"]
        print(config)
        for a, v in config.items():
            print(a)
            setattr(self, a, v)
        if len(required_attributes) > 0:
            has_required = [hasattr(self, a) for a in required_attributes]
            if not all(has_required):
                for i in range(len(required_attributes)):
                    if not has_required[i]:
                        print("{0} not found.".format(required_attributes[i]))
                        exit(1)

